import difflib
import os
import re
from collections import defaultdict
from pathlib import Path

from check.focs_dump_parser._tools import get_base_dump_folder, get_branch

_negative_number = re.compile(r"\(-\(?\d+\.?\d*\)\)")


def read_file(path: Path):
    if not path.exists():
        return []
    else:
        with path.open() as f:
            return f.readlines()


def prettify_dump(text):
    text = text.replace("(0 = LocalCandidate.TurnsSinceFocusChange)", "(LocalCandidate.TurnsSinceFocusChange = 0)")

    def replacer(match):
        return match.group(0).replace("(", "").replace(")", "")

    return _negative_number.sub(replacer, text)


def report_diff():
    base = get_base_dump_folder()

    a_branch = "master"

    b_branch = get_branch()

    path_a = base / a_branch
    path_b = base / b_branch

    files_a = os.listdir(path_a)
    files_b = os.listdir(path_b)

    all_files = {*files_a, *files_b}

    clean_species = []

    stats = defaultdict(list)

    for name in all_files:
        content_a = read_file(path_a / name)
        content_b = read_file(path_b / name)

        name = name.removesuffix(".txt")

        a = [prettify_dump(x) for x in content_a]
        b = [prettify_dump(x) for x in content_b]

        if a == b == []:
            stats["both empty"].append(name)
        if a == b:
            stats["same"].append(name)
        elif content_a == []:
            stats["file mismatch"].append(f"{name}.{a_branch} is missed")
        elif content_b == []:
            stats["file mismatch"].append(f"{name}.{b_branch} is missed")

        else:
            stats["diff"].append(name)
            for line in difflib.unified_diff(a, b, f"{name}.{a_branch}", f"{name}.{b_branch}"):
                print(line, end="")

    print(clean_species)
    print(f"Finish comparing {len(all_files)} files")
    return stats


if __name__ == "__main__":
    report_diff()



from macros.base_prod import RESEARCH_PER_POP
from macros.priorities import (
    TARGET_AFTER_SCALING_PRIORITY,
    TARGET_EARLY_BEFORE_SCALING_PRIORITY,
    TARGET_SCALING_PRIORITY,
)

try:
    from focs._effects import (
        EffectsGroup,
        Focus,
        GalaxyMaxAIAggression,
        Happiness,
        IsHuman,
        IsSource,
        Planet,
        SetTargetResearch,
        Target,
        Value,
    )
except ModuleNotFoundError:
    pass

from macros.multiplier import BAD_MULTIPLIER

# NO_RESEARCH
# '''[[DESCRIPTION_EFFECTSGROUP_MACRO(NO_RESEARCH_DESC)]]'''

BASIC_RESEARCH = [
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Focus(type=["FOCUS_RESEARCH"]) & Happiness(low=0),
        accountinglabel="FOCUS_RESEARCH_LABEL",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        effects=SetTargetResearch(value=Value + Target.Population * RESEARCH_PER_POP),
    ),
    EffectsGroup(  # gives human bonuses when AI Aggression set to Beginner
        scope=IsSource,
        activation=Planet() & IsHuman & (GalaxyMaxAIAggression == 0),  # human player, not human species
        accountinglabel="DIFFICULTY",
        priority=TARGET_AFTER_SCALING_PRIORITY,
        effects=SetTargetResearch(value=Value * 2),
    ),
]

# VERY_BAD_RESEARCH
# '''[[BASIC_RESEARCH]]

#         EffectsGroup
#             description = "VERY_BAD_RESEARCH_DESC"
#             scope = Source
#             activation = And [
#                 Planet
#                 Focus type = "FOCUS_RESEARCH"
#             ]
#             accountinglabel = "VERY_BAD_RESEARCH_LABEL"
#             priority = [[TARGET_SCALING_PRIORITY]]
#             effects = SetTargetResearch value = Value*[[VERY_BAD_MULTIPLIER]]
# '''

BAD_RESEARCH = [
    *BASIC_RESEARCH,
    EffectsGroup(
        description="BAD_RESEARCH_DESC",
        scope=IsSource,
        activation=Planet() & Focus(type=["FOCUS_RESEARCH"]),
        accountinglabel="BAD_RESEARCH_LABEL",
        priority=TARGET_SCALING_PRIORITY,
        effects=SetTargetResearch(value=Value * BAD_MULTIPLIER),
    ),
]

# AVERAGE_RESEARCH
# '''[[BASIC_RESEARCH]]
# '''

# GOOD_RESEARCH
# '''[[BASIC_RESEARCH]]

#         EffectsGroup
#             description = "GOOD_RESEARCH_DESC"
#             scope = Source
#             activation = And [
#                 Planet
#                 Focus type = "FOCUS_RESEARCH"
#                 Happiness low = 2
#             ]
#             accountinglabel = "GOOD_RESEARCH_LABEL"
#             priority = [[TARGET_SCALING_PRIORITY]]
#             effects = SetTargetResearch value = Value*[[GOOD_MULTIPLIER]]
# '''

# GREAT_RESEARCH
# '''[[BASIC_RESEARCH]]

#         EffectsGroup
#             description = "GREAT_RESEARCH_DESC"
#             scope = Source
#             activation = And [
#                 Planet
#                 Focus type = "FOCUS_RESEARCH"
#                 Happiness low = 0
#             ]
#             accountinglabel = "GREAT_RESEARCH_LABEL"
#             priority = [[TARGET_SCALING_PRIORITY]]
#             effects = SetTargetResearch value = Value*[[GREAT_MULTIPLIER]]
# '''

# ULTIMATE_RESEARCH
# '''[[BASIC_RESEARCH]]

#         EffectsGroup
#             description = "ULTIMATE_RESEARCH_DESC"
#             scope = Source
#             activation = And [
#                 Planet
#                 Focus type = "FOCUS_RESEARCH"
#             ]
#             accountinglabel = "ULTIMATE_RESEARCH_LABEL"
#             priority = [[TARGET_SCALING_PRIORITY]]
#             effects = SetTargetResearch value = Value*[[ULTIMATE_MULTIPLIER]]
# '''



try:
    from focs._effects import EffectsGroup, GameRule, IsSource, NoEffect
except ModuleNotFoundError:
    pass


MIN_RECOLONIZING_SIZE = 3

MIN_RECOLONIZING_HAPPINESS = 5

IMPOSSIBLY_LARGE_TURN = 2**15


def DESCRIPTION_EFFECTSGROUP_MACRO(desc: str):
    return EffectsGroup(description=desc, scope=IsSource, activation=None, effects=NoEffect)


FIGHTER_DAMAGE_FACTOR = GameRule(type=float, name="RULE_FIGHTER_DAMAGE_FACTOR")

PLANET_DEFENSE_FACTOR = GameRule(type=float, name="RULE_SHIP_WEAPON_DAMAGE_FACTOR")

PLANET_SHIELD_FACTOR = GameRule(type=float, name="RULE_SHIP_STRUCTURE_FACTOR")

SHIP_WEAPON_DAMAGE_FACTOR = GameRule(type=float, name="RULE_SHIP_WEAPON_DAMAGE_FACTOR")

SHIP_SHIELD_FACTOR = GameRule(type=float, name="RULE_SHIP_WEAPON_DAMAGE_FACTOR")

SHIP_STRUCTURE_FACTOR = GameRule(type=float, name="RULE_SHIP_STRUCTURE_FACTOR")

SYSTEM_MINES_DAMAGE_FACTOR = GameRule(type=float, name="RULE_SHIP_STRUCTURE_FACTOR")

SUPPLY_DISCONNECTED_INFLUENCE_MALUS = 1


# empire id used for unowned planets/ships - as defined in Universe.cpp(?)
UNOWNED_EMPIRE_ID = -1



GameRule(
    name="RULE_SHIP_HULL_COST_FACTOR",
    description="RULE_SHIP_HULL_COST_FACTOR_DESC",
    category="BALANCE",
    type=float,
    default=1.0,
    min=0.1,
    max=10.0,
    rank=20,
)

GameRule(
    name="RULE_SHIP_PART_COST_FACTOR",
    description="RULE_SHIP_PART_COST_FACTOR_DESC",
    category="BALANCE",
    type=float,
    default=1.0,
    min=0.1,
    max=10.0,
    rank=10,
)

GameRule(
    name="RULE_TECH_COST_FACTOR",
    description="RULE_TECH_COST_FACTOR_DESC",
    category="BALANCE",
    type=float,
    default=2.0,
    min=0.1,
    max=10.0,
    rank=666,
)

GameRule(
    name="RULE_BUILDING_COST_FACTOR",
    description="RULE_BUILDING_COST_FACTOR_DESC",
    category="BALANCE",
    type=float,
    default=1.0,
    min=0.1,
    max=10.0,
    rank=1337,
)

GameRule(
    name="RULE_ENABLE_EXPERIMENTORS",
    description="RULE_ENABLE_EXPERIMENTORS_DESC",
    category="CONTENT",
    type=bool,
    default=True,
    rank=1,
)

GameRule(
    name="RULE_ENABLE_SUPER_TESTER",
    description="RULE_ENABLE_SUPER_TESTER_DESC",
    category="CONTENT",
    type=bool,
    default=True,
    rank=2,
)

GameRule(
    name="RULE_TEST_STRING",
    description="RULE_TEST_STRING_DESC",
    category="TEST",
    type=str,
    default="PLAYER",
    allowed=["MODERATOR", "OBSERVER", "PLAYER", "AI_PLAYER"],
    rank=999,
)

GameRule(
    name="RULE_HABITABLE_SIZE_TINY",
    description="RULE_HABITABLE_SIZE_DESC",
    category="PLANET_SIZE",
    type=int,
    default=1,
    min=0,
    max=999,
    rank=900000,
)

GameRule(
    name="RULE_HABITABLE_SIZE_SMALL",
    description="RULE_HABITABLE_SIZE_DESC",
    category="PLANET_SIZE",
    type=int,
    default=2,
    min=0,
    max=999,
    rank=900001,
)

GameRule(
    name="RULE_HABITABLE_SIZE_MEDIUM",
    description="RULE_HABITABLE_SIZE_DESC",
    category="PLANET_SIZE",
    type=int,
    default=3,
    min=0,
    max=999,
    rank=900002,
)

GameRule(
    name="RULE_HABITABLE_SIZE_LARGE",
    description="RULE_HABITABLE_SIZE_DESC",
    category="PLANET_SIZE",
    type=int,
    default=4,
    min=0,
    max=999,
    rank=900003,
)

GameRule(
    name="RULE_HABITABLE_SIZE_HUGE",
    description="RULE_HABITABLE_SIZE_DESC",
    category="PLANET_SIZE",
    type=int,
    default=5,
    min=0,
    max=999,
    rank=900004,
)

GameRule(
    name="RULE_HABITABLE_SIZE_ASTEROIDS",
    description="RULE_HABITABLE_SIZE_DESC",
    category="PLANET_SIZE",
    type=int,
    default=3,
    min=0,
    max=999,
    rank=900005,
)

GameRule(
    name="RULE_HABITABLE_SIZE_GASGIANT",
    description="RULE_HABITABLE_SIZE_DESC",
    category="PLANET_SIZE",
    type=int,
    default=6,
    min=0,
    max=999,
    rank=900006,
)

GameRule(
    name="RULE_SHIP_PART_BASED_UPKEEP",
    description="RULE_SHIP_PART_BASED_UPKEEP_DESC",
    category="BALANCE",
    type=bool,
    default=False,
    rank=1234567,
)

GameRule(
    name="RULE_ENABLE_ALLIED_REPAIR",
    description="RULE_ENABLE_ALLIED_REPAIR_DESC",
    category="MULTIPLAYER",
    type=bool,
    default=False,
    rank=3001,
)

GameRule(
    name="RULE_FIRST_COMBAT_ROUND_IN_CLOSE_TARGETING_RANGE",
    description="RULE_FIRST_COMBAT_ROUND_IN_CLOSE_TARGETING_RANGE_DESC",
    category="BALANCE",
    type=int,
    default=3,
    min=1,
    max=20,
    rank=42,
)



from macros.priorities import AFTER_ALL_TARGET_MAX_METERS_PRIORITY

# NO_SUPPLY
# '''
# '''

try:
    from focs._effects import (
        CurrentTurn,
        EffectsGroup,
        IsSource,
        LocalCandidate,
        MinOf,
        Planet,
        SetMaxSupply,
        SetSupply,
        Target,
        Value,
    )
except ModuleNotFoundError:
    pass


STANDARD_SUPPLY_GROWTH = EffectsGroup(  # increase 1 per turn, up to max
    scope=IsSource,
    activation=Planet()
    & (LocalCandidate.LastTurnConquered < CurrentTurn)
    & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn),
    priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
    effects=SetSupply(value=MinOf(float, Value(Target.MaxSupply), Value + 1)),
)


# VERY_BAD_SUPPLY
# '''     EffectsGroup
#             description = "VERY_BAD_SUPPLY_DESC"
#             scope = Source
#             activation = Planet
#             accountinglabel = "VERY_BAD_SUPPLY_LABEL"
#             effects = SetMaxSupply value = Value - 1

# [[STANDARD_SUPPLY_GROWTH]]
# '''

# BAD_SUPPLY
# '''     EffectsGroup
#             description = "BAD_SUPPLY_DESC"
#             scope = Source
#             activation = Planet
#             accountinglabel = "BAD_SUPPLY_LABEL"
#             effects = SetMaxSupply value = Value + 0

# [[STANDARD_SUPPLY_GROWTH]]
# '''

AVERAGE_SUPPLY = [
    EffectsGroup(
        scope=IsSource,
        activation=Planet(),
        accountinglabel="AVERAGE_SUPPLY_LABEL",
        effects=SetMaxSupply(value=Value + 1),
    ),
    STANDARD_SUPPLY_GROWTH,
]

# GREAT_SUPPLY
# '''     EffectsGroup
#             description = "GREAT_SUPPLY_DESC"
#             scope = Source
#             activation = Planet
#             accountinglabel = "GREAT_SUPPLY_LABEL"
#             effects = SetMaxSupply value = Value + 2

# [[STANDARD_SUPPLY_GROWTH]]
# '''

# ULTIMATE_SUPPLY
# '''     EffectsGroup
#             description = "ULTIMATE_SUPPLY_DESC"
#             scope = Source
#             activation = Planet
#             accountinglabel = "ULTIMATE_SUPPLY_LABEL"
#             effects = SetMaxSupply value = Value + 3
#
# [[STANDARD_SUPPLY_GROWTH]]
# '''



import argparse
import fnmatch
import os
from collections import defaultdict

arguments = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description="Parses files for content tags and creates and a doxygen suitable listing from the parse results.",
    epilog="%(prog)s Parses FILES in the MATCH directory for TAGs."
    "The comment lines may be in either # or /// form in a file,"
    " but only one form throughout a file.  The results of the parse are"
    ' formatted for a doxygen ".dox" file.',
)
arguments.add_argument("--match", default="./", help="Root directory to search for files to parse.")
arguments.add_argument(
    "-f", "--files", metavar="MASK", nargs="*", default=["*.*"], help="One or more file masks of files to parse."
)
arguments.add_argument(
    "-x",
    "--exclude",
    metavar="MASK",
    nargs="*",
    default=[],
    help="One or more file masks to exclude from results matching source_mask(s).",
)
arguments.add_argument(
    "--tag",
    nargs=2,
    default=["@content_tag{", "}"],
    help="Set of opening and closing" " strings which surround a tag name and precede the tag documentation.",
)
arguments.add_argument("-o", "--output", default="output.txt", help="Full path to an output file.")
arguments.add_argument(
    "--link_source",
    metavar="CONTEXT_DIR BASE_URL LINE_PREFIX]",
    nargs="*",
    default=[],
    help="Display linked source before each tag documentation. Requires arguments in order:"
    " CONTEXT_DIR BASE_URL LINE_PREFIX. File sources will link to"
    " {BASE_URL}{FILE}{LINE_PREFIX}{LINE_NUM}, where FILE is the relative path of CONTEXT_DIR."
    " e.g. Given: FILE=a/b/c/xyz.txt LINE_NUM=4 CONTEXT_DIR=a/b BASE_URL=http://a.it LINE_PREFIX=?"
    " Link result will be: http://a.it/c/xyz.txt?4",
)
arguments.add_argument(
    "--dry_run", action="store_true", help="Does not write to OUTPUT, instead printing info" " to stdout."
)

args = vars(arguments.parse_args())

tag_open = args.get("tag")[0]
tag_close = args.get("tag")[1]


def add_doc_source(_file_name, _line_number, _content, _tags):
    fn = _file_name.lstrip("./")
    # index documentation by tag name storing the source file name, line number, and documentation string
    _tags[_content[0]].add((fn, _line_number, _content[1].strip()))


def parse_file(_parse_file, _tags):  # noqa: C901
    with open(_parse_file, encoding="utf-8") as f:
        match_line = 0
        content = []
        special_comments_this_file = None
        # Documentation may span multiple lines for a content tag.
        # Matched lines are only stored in _tags once an end condition is met.
        # End conditions are: a line not starting a comment, a new content tag, end of file
        for raw_line in enumerate(f):
            if raw_line[1].lstrip().startswith("#") and special_comments_this_file is not True:
                if special_comments_this_file is None:
                    special_comments_this_file = False
                if tag_open in raw_line[1]:
                    if match_line:
                        # in event focs scripts decide to stack one documentation on top of another
                        add_doc_source(_parse_file, match_line, content, _tags)
                    # store content and line for later addition
                    content = "".join(raw_line[1].split(tag_open, 1)[1]).split(tag_close, 1)
                    content[1] = content[1].strip()
                    match_line = raw_line[0] + 1
                elif match_line:
                    # not a new content tag, append to previous line description
                    content[1] += " " + raw_line[1].lstrip("# ")
            elif raw_line[1].lstrip().startswith("/// ") and special_comments_this_file is not False:
                if special_comments_this_file is None:
                    special_comments_this_file = True
                    content = ["# "]
                if tag_open in raw_line[1]:
                    if match_line:
                        # in event focs scripts decide to stack one documentation on top of another
                        add_doc_source(_parse_file, match_line, content, _tags)
                    # store content and line for later addition
                    content = "".join(raw_line[1].split(tag_open, 1)[1]).split(tag_close, 1)
                    content[1] = content[1].strip()
                    match_line = raw_line[0] + 1
                elif match_line:
                    # not a new content tag, append to previous line description
                    content[1] += " " + raw_line[1].lstrip("/// ")
            elif match_line:
                # end of description, add a node for this source
                add_doc_source(_parse_file, match_line, content, _tags)
                match_line = 0
                content = []
        # EOF
        if match_line:
            # in event a matched comment is the last line in a file
            add_doc_source(_parse_file, match_line, content, _tags)


def get_link(_file, _line_number):
    link_args = args.get("link_source")
    if len(link_args) != 3:
        if link_args:
            print("Invalid link_source arguments: ", link_args)
        return ""
    base_path = os.path.relpath(_file, link_args[0].lstrip("./"))
    link = os.path.join(link_args[1], base_path)
    link += str.format("{0}{1}", link_args[2], _line_number)
    return str.format("[{0}]({1})", base_path, link)


all_tags = defaultdict(set)

# parse files for tags
for _file_path, _, _files in os.walk(args.get("match")):
    for file_name in enumerate(_files):
        file_match = False
        # if this file should be checked
        for mask in args.get("files"):
            if fnmatch.fnmatch(file_name[1], mask):
                file_match = True
        if file_match:
            # check for exclusions
            for mask in args.get("exclude"):
                if fnmatch.fnmatch(file_name[1], mask):
                    file_match = False
            if file_match:
                parse_file(os.path.join(_file_path, file_name[1]), all_tags)

output_buff = []

# create tags formatted for doxygen

output_buff.append("/**\n")
output_buff.append("@page content_tag_listing Content Definition Tag Listing\n")
output_buff.append(
    "This page is required to load parsed data into the documentation, the actual listing can be found @ref content_tags\n"
)
for tag in sorted(all_tags):
    output_buff.append("@content_tag{" + tag + "} ")
    for source in all_tags[tag]:
        description = str.format("{0}  ", source[2])
        # prepend source and link if requested
        link_str = get_link(source[0], source[1])
        if link_str:
            link_str += " - "
        output_buff.append(link_str + description)
    output_buff.append("\n")
output_buff.append("*/\n")

if args.get("dry_run"):
    source_count = 0
    for tag in all_tags:
        for value in all_tags[tag]:
            source_count += 1
    print("arguments:")
    for arg in args:
        print("\t", arg, ":", args[arg])
    print(f"\nFound {len(all_tags.keys())} tags with {source_count} total sources")
else:
    abs_dir = os.path.dirname(os.path.abspath(args.get("output")))
    if not os.path.exists(abs_dir):
        os.mkdir(abs_dir)
    with open(args.get("output"), "w") as output_file:
        output_file.writelines(output_buff)



from macros.base_prod import TROOPS_PER_POP
from macros.priorities import (
    AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
    TARGET_AFTER_2ND_SCALING_PRIORITY,
    TARGET_EARLY_BEFORE_SCALING_PRIORITY,
    TARGET_SCALING_PRIORITY,
)

try:
    from focs._effects import (
        Abs,
        Adequate,
        Capital,
        CurrentTurn,
        EffectsGroup,
        EmpireStockpile,
        Focus,
        GalaxyMaxAIAggression,
        Happiness,
        HasEmpireStockpile,
        HasSpecies,
        Homeworld,
        Hostile,
        Influence,
        IsHuman,
        IsSource,
        LocalCandidate,
        MaxOf,
        MinOf,
        NamedReal,
        OwnedBy,
        Planet,
        Poor,
        SetMaxTroops,
        SetRebelTroops,
        SetTargetHappiness,
        SetTargetInfluence,
        SetTroops,
        Source,
        StatisticCount,
        Target,
        Unowned,
        Value,
    )
except ModuleNotFoundError:
    pass


UNSTABLE_REBEL_TROOPS = [
    EffectsGroup(
        scope=IsSource,
        activation=~Unowned & ~Happiness(low=0),
        accountinglabel="INSTABILITY_REBELLION",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        # more negative stability and more population generates more rebels each turn
        # TODO: scale with some species trait?
        effects=SetRebelTroops(value=Value + Abs(float, Target.Happiness) * (Target.Population / 5.0) ** 0.5),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & ~Unowned & ~HasEmpireStockpile(empire=Source.Owner, resource=Influence, low=0) & ~Capital
        # TODO: some species trait?
        ,
        accountinglabel="PARTISANS_REBELLION",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        effects=SetTargetHappiness(
            value=Value - 0.5 * Abs(float, EmpireStockpile(empire=Source.Owner, resource=Influence)) ** 0.5
        ),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & ~Unowned & ~HasEmpireStockpile(empire=Source.Owner, resource=Influence, low=0) & ~Capital,
        accountinglabel="INFLUENCE_DEBT_DECAY",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        effects=SetTargetInfluence(
            value=Value
            + 0.5
            * Abs(float, EmpireStockpile(empire=Source.Owner, resource=Influence)) ** 0.5
            / MaxOf(float, 1.0, StatisticCount(float, condition=Planet() & HasSpecies() & OwnedBy(empire=Source.Owner)))
        ),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet(environment=[Adequate]),
        accountinglabel="DEPENDENCE_ON_IMPERIAL_SUPPORT_DUE_TO_ENVIRONMENT",
        priority=TARGET_SCALING_PRIORITY,
        effects=SetRebelTroops(value=MaxOf(float, 0.0, Abs(float, Value / 1.5))),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet(environment=[Poor]),
        accountinglabel="DEPENDENCE_ON_IMPERIAL_SUPPORT_DUE_TO_ENVIRONMENT",
        priority=TARGET_SCALING_PRIORITY,
        effects=SetRebelTroops(value=MaxOf(float, 0.0, Abs(float, Value / 2.0))),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet(environment=[Hostile]),
        accountinglabel="DEPENDENCE_ON_IMPERIAL_SUPPORT_DUE_TO_ENVIRONMENT",
        priority=TARGET_SCALING_PRIORITY,
        effects=SetRebelTroops(value=MaxOf(float, 0.0, Abs(float, Value / 4.0))),
    ),
]

BASIC_DEFENSE_TROOPS = [
    EffectsGroup(
        scope=IsSource,
        activation=Homeworld() & ~Unowned,
        stackinggroup="HOMEWORLD_TROOPS_STACK",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        effects=SetMaxTroops(value=Value + 4),
        # accountinglabel="HOMEWORLD_LABEL",
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & ~Unowned,
        accountinglabel="IMPERIAL_GARRISON_LABEL",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        effects=SetMaxTroops(value=Value + 6),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Unowned,
        stackinggroup="BASIC_TROOPS_STACK",
        accountinglabel="INDEPENDENT_TROOP_LABEL",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        effects=SetMaxTroops(value=Value + 10),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet(),
        stackinggroup="POPULATION_TROOPS_STACK",
        accountinglabel="DEF_ROOT_DEFENSE",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        effects=SetMaxTroops(
            value=Value
            + Target.Population * NamedReal(name="BASIC_DEFENSE_TROOPS_MAX_TROOPS_PERPOP", value=1.0 * TROOPS_PER_POP)
        ),
    ),
    EffectsGroup(  # gives human bonuses when AI Aggression set to Beginner
        scope=IsSource,
        activation=IsHuman & (GalaxyMaxAIAggression == 0) & Planet(),  # human player, not human species
        accountinglabel="DIFFICULTY",
        effects=SetMaxTroops(value=MaxOf(float, 6, Value * 2)),
    ),
    EffectsGroup(  # base troops regeneration
        scope=IsSource,
        activation=Planet()
        & (LocalCandidate.LastTurnConquered < CurrentTurn)
        & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn),
        priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
        effects=SetTroops(
            value=MinOf(
                float, Value(Target.MaxTroops), Value + NamedReal(name="BASIC_DEFENSE_TROOPS_TROOPREGEN_FLAT", value=1)
            )
        ),
    ),
]

PROTECTION_FOCUS_TROOPS = [
    EffectsGroup(  # double max troops
        scope=IsSource,
        activation=Planet() & Focus(type=["FOCUS_PROTECTION"]),
        stackinggroup="FOCUS_PROTECTION_TROOPS_STACK",
        priority=TARGET_AFTER_2ND_SCALING_PRIORITY,
        effects=SetMaxTroops(value=MaxOf(float, Value, Value * 2)),
        # accountinglabel="FOCUS_PROTECTION_LABEL",
    ),
    EffectsGroup(  # increase troop growth rate
        scope=IsSource,
        activation=Planet()
        & Focus(type=["FOCUS_PROTECTION"])
        & (LocalCandidate.LastTurnConquered < CurrentTurn)
        & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn),
        accountinglabel="FOCUS_PROTECTION_LABEL",
        priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
        effects=SetTroops(
            value=MinOf(
                float, Value(Target.MaxTroops), Value + MaxOf(float, Target.MaxTroops, Target.Population) ** 0.5
            )
        ),
    ),
]


# NO_DEFENSE_TROOPS
# '''    EffectsGroup
#         description = "NO_DEFENSE_TROOPS_DESC"
#         scope = Source
#         activation = None
#         effects = SetMaxTroops value = 0

# [[UNSTABLE_REBEL_TROOPS]]
# '''

BAD_DEFENSE_TROOPS = [
    *BASIC_DEFENSE_TROOPS,
    EffectsGroup(
        description="BAD_DEFENSE_TROOPS_DESC",
        scope=IsSource,
        activation=Planet(),
        effects=SetMaxTroops(value=Value * 0.5),
        # accountinglabel="BAD_TROOPS_LABEL",
    ),
    *UNSTABLE_REBEL_TROOPS,
    *PROTECTION_FOCUS_TROOPS,
]

# AVERAGE_DEFENSE_TROOPS
# '''[[BASIC_DEFENSE_TROOPS]]

# [[UNSTABLE_REBEL_TROOPS]]

# [[PROTECTION_FOCUS_TROOPS]]
# '''

# GOOD_DEFENSE_TROOPS
# '''[[BASIC_DEFENSE_TROOPS]]

#     EffectsGroup
#         description = "GOOD_DEFENSE_TROOPS_DESC"
#         scope = Source
#         activation = Planet
#         effects = SetMaxTroops value = Value * 1.5
#         accountinglabel = "GOOD_TROOPS_LABEL"

# [[UNSTABLE_REBEL_TROOPS]]

# [[PROTECTION_FOCUS_TROOPS]]
# '''

# GREAT_DEFENSE_TROOPS
# '''[[BASIC_DEFENSE_TROOPS]]

#     EffectsGroup
#         description = "GREAT_DEFENSE_TROOPS_DESC"
#         scope = Source
#         activation = Planet
#         effects = SetMaxTroops value = Value * 2
#         accountinglabel = "GREAT_TROOPS_LABEL"

# [[UNSTABLE_REBEL_TROOPS]]

# [[PROTECTION_FOCUS_TROOPS]]
# '''

# ULTIMATE_DEFENSE_TROOPS
# '''[[BASIC_DEFENSE_TROOPS]]

#     EffectsGroup
#         description = "ULTIMATE_DEFENSE_TROOPS_DESC"
#         scope = Source
#         activation = Planet
#         effects = SetMaxTroops value = Value * 3
#         accountinglabel = "ULTIMATE_TROOPS_LABEL"

# [[UNSTABLE_REBEL_TROOPS]]

# [[PROTECTION_FOCUS_TROOPS]]
# '''

# NO_OFFENSE_TROOPS
# '''EffectsGroup
#             description = "NO_OFFENSE_TROOPS_DESC"
#             scope = Source
#             activation = And [
#                 Ship
#                 Or [
#                     DesignHasPart name = "GT_TROOP_POD"
#                     DesignHasPart name = "GT_TROOP_POD_2"
#                 ]
#             ]
#             stackinggroup = "NO_OFFENSIVE_TROOPS_STACK"
#             accountinglabel = "NO_OFFENSIVE_TROOPS_LABEL"
#             effects = [
#                 SetCapacity partname = "GT_TROOP_POD"   value = 0
#                 SetCapacity partname = "GT_TROOP_POD_2" value = 0
#             ]
# '''

# BAD_OFFENSE_TROOPS
# '''EffectsGroup
#             description = "BAD_OFFENSE_TROOPS_DESC"
#             scope = Source
#             activation = And [
#                 Ship
#                 Or [
#                     DesignHasPart name = "GT_TROOP_POD"
#                     DesignHasPart name = "GT_TROOP_POD_2"
#                 ]
#             ]
#             stackinggroup = "BAD_OFFENSIVE_TROOPS_STACK"
#             accountinglabel = "BAD_OFFENSIVE_TROOPS_LABEL"
#             effects = [
#                 SetCapacity partname = "GT_TROOP_POD"   value = (PartCapacity name = "GT_TROOP_POD") * 0.5
#                 SetCapacity partname = "GT_TROOP_POD_2" value = (PartCapacity name = "GT_TROOP_POD_2") * 0.5
#             ]
# '''

# //AVERAGE_OFFENSE_TROOPS has no effects

# GOOD_OFFENSE_TROOPS
# '''EffectsGroup
#             description = "GOOD_OFFENSE_TROOPS_DESC"
#             scope = Source
#             activation = And [
#                 Ship
#                 Or [
#                     DesignHasPart name = "GT_TROOP_POD"
#                     DesignHasPart name = "GT_TROOP_POD_2"
#                 ]
#             ]
#             stackinggroup = "GOOD_OFFENSIVE_TROOPS_STACK"
#             accountinglabel = "GOOD_OFFENSIVE_TROOPS_LABEL"
#             effects = [
#                 SetCapacity partname = "GT_TROOP_POD"   value = (PartCapacity name = "GT_TROOP_POD") * 1.5
#                 SetCapacity partname = "GT_TROOP_POD_2" value = (PartCapacity name = "GT_TROOP_POD_2") * 1.5
#             ]
# '''

# GREAT_OFFENSE_TROOPS
# '''EffectsGroup
#             description = "GREAT_OFFENSE_TROOPS_DESC"
#             scope = Source
#             activation = And [
#                 Ship
#                 Or [
#                     DesignHasPart name = "GT_TROOP_POD"
#                     DesignHasPart name = "GT_TROOP_POD_2"
#                 ]
#             ]
#             stackinggroup = "GREAT_OFFENSIVE_TROOPS_STACK"
#             accountinglabel = "GREAT_OFFENSIVE_TROOPS_LABEL"
#             effects = [
#                 SetCapacity partname = "GT_TROOP_POD"   value = (PartCapacity name = "GT_TROOP_POD") * 2
#                 SetCapacity partname = "GT_TROOP_POD_2" value = (PartCapacity name = "GT_TROOP_POD_2") * 2
#             ]
# '''

# ULTIMATE_OFFENSE_TROOPS
# '''EffectsGroup
#             description = "ULTIMATE_OFFENSE_TROOPS_DESC"
#             scope = Source
#             activation = And [
#                 Ship
#                 Or [
#                     DesignHasPart name = "GT_TROOP_POD"
#                     DesignHasPart name = "GT_TROOP_POD_2"
#                 ]
#             ]
#             stackinggroup = "ULTIMATE_OFFENSIVE_TROOPS_STACK"
#             accountinglabel = "ULTIMATE_OFFENSIVE_TROOPS_LABEL"
#             effects = [
#                 SetCapacity partname = "GT_TROOP_POD"   value = (PartCapacity name = "GT_TROOP_POD") * 3
#                 SetCapacity partname = "GT_TROOP_POD_2" value = (PartCapacity name = "GT_TROOP_POD_2") * 3
#             ]
# '''



from macros.misc import PLANET_SHIELD_FACTOR
from macros.priorities import (
    AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
    TARGET_AFTER_2ND_SCALING_PRIORITY,
)

# // param1: value
# NATIVE_PLANETARY_SHIELDS
# '''EffectsGroup
#             scope = Source
#             activation = And [
#                 Planet
#                 Unowned
#             ]
#             accountinglabel = "NATIVE_PLANETARY_SHIELDS_LABEL"
#             effects = SetMaxShield value = Value + ( @1@ )
# '''

try:
    from focs._effects import (
        CurrentTurn,
        EffectsGroup,
        Focus,
        IsSource,
        LocalCandidate,
        MinOf,
        Planet,
        SetMaxShield,
        SetShield,
        Target,
        Unowned,
        Value,
    )
except ModuleNotFoundError:
    pass

STANDARD_SHIELD_GROWTH = EffectsGroup(  # increase 1 per turn, up to max
    scope=IsSource,
    activation=Planet()
    & Unowned
    & (LocalCandidate.LastTurnConquered < CurrentTurn)
    & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn),
    accountinglabel="DEF_ROOT_DEFENSE",
    priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
    effects=SetShield(value=MinOf(float, Value(Target.MaxShield), Value + PLANET_SHIELD_FACTOR)),
)


PROTECTION_FOCUS_SHIELDS = EffectsGroup(
    scope=IsSource,
    activation=Planet() & Focus(type=["FOCUS_PROTECTION"]),
    stackinggroup="FOCUS_PROTECTION_SHIELDS_STACK",
    accountinglabel="FOCUS_PROTECTION_LABEL",
    priority=TARGET_AFTER_2ND_SCALING_PRIORITY,
    effects=SetMaxShield(value=Value * 2),
)


AVERAGE_PLANETARY_SHIELDS = [PROTECTION_FOCUS_SHIELDS, STANDARD_SHIELD_GROWTH]

# #include "../macros/priorities.macros"
# #include "../macros/misc.macros"



#!/usr/bin/python3

import pathlib
import re

MIN_CPP_INCLUDE_COUNT = 5
EXCLUDE_LIBS = ["GG"]
IGNORE_INCLUDES = ["StdAfx.h", "stdafx.h"]

PREAMBLE = """
#pragma once

// We include all external headers used in any of the header files,
// plus external headers used in at least five .cpp files.

// https://hownot2code.com/2016/08/16/stdafx-h/

"""

POSTAMBLE = """

#ifdef _MSC_VER
// Note: This is a workaround for Visual C++ non-conformant pre-processor
// handling of empty macro arguments.
#include <boost/preprocessor/control/if.hpp>
#include <boost/preprocessor/facilities/is_empty_variadic.hpp>
#endif
"""


def find_files_from_vcxproj(vcxproj, includetype, fileextension):
    pattern = r"^\s*<" + includetype + r"\s+Include=\"(?P<include>[^\"]+" + fileextension + r")\"\s*/>"
    vcxproj_pattern = re.compile(pattern)
    base_path = pathlib.Path(vcxproj).parent

    with open(vcxproj) as f:
        for i, line in enumerate(f):
            match = vcxproj_pattern.match(line)
            if match and match.group("include") not in IGNORE_INCLUDES:
                in_project_path = pathlib.PureWindowsPath(match.group("include"))
                yield str(base_path.joinpath(in_project_path))


include_pattern = re.compile(r"^#include[ \t]+<(?P<include>(?:(?P<lib>[^/>]+)/)?[^>]+)>")
ifdef_pattern = re.compile(r"^\s*#if(?:def)?\s")
endif_pattern = re.compile(r"^\s*#endif")


def find_includes_from_cpp_or_h(filename):
    if_nesting_level = 0
    with open(filename) as f:
        for i, line in enumerate(f):
            if ifdef_pattern.match(line):
                if_nesting_level += 1
            elif endif_pattern.match(line):
                if_nesting_level -= 1
            elif if_nesting_level == 0:
                match = include_pattern.match(line)
                if match and match.group("include") not in IGNORE_INCLUDES:
                    lib = ""
                    if match.group("lib"):
                        lib = match.group("lib")
                    yield (match.group("include"), lib)


def assemble_precompile_header_includes(headeronly_vcxproj_files, vcxproj_files):
    headers_includes = {}
    cpp_includes = {}

    index = -1
    for vcxproj in headeronly_vcxproj_files + vcxproj_files:
        index = index + 1
        for include_file in find_files_from_vcxproj(vcxproj, "ClInclude", ".h"):
            for include, lib in find_includes_from_cpp_or_h(include_file):
                if include in headers_includes:
                    continue
                headers_includes[include] = (index, vcxproj, lib)

    index = -1
    for vcxproj in vcxproj_files:
        index = index + 1
        for include_file in find_files_from_vcxproj(vcxproj, "ClCompile", ".cpp"):
            for include, lib in find_includes_from_cpp_or_h(include_file):
                if include not in cpp_includes:
                    cpp_includes[include] = [index, vcxproj, lib, 1]
                else:
                    cpp_includes[include][3] = cpp_includes[include][3] + 1

    headers = [
        (
            "includes from .h",
            headers_includes[include][0],
            headers_includes[include][1],
            headers_includes[include][2],
            include,
            0,
        )
        for include in headers_includes
    ]
    headers.sort()

    cpps = [
        (
            "includes from .cpp",
            cpp_includes[include][0],
            cpp_includes[include][1],
            cpp_includes[include][2],
            include,
            cpp_includes[include][3],
        )
        for include in cpp_includes
        if include not in headers_includes and cpp_includes[include][3] >= MIN_CPP_INCLUDE_COUNT
    ]
    cpps.sort()

    return headers + cpps


def update_stdafx_h(stdafx_h_filename, header_only_vcxproj_files, vcxproj_files):
    last_type = ""
    last_proj = ""
    last_lib = None

    lines = []

    for type, index, proj, lib, include, count in assemble_precompile_header_includes(
        header_only_vcxproj_files, vcxproj_files
    ):
        if lib in EXCLUDE_LIBS:
            continue
        if type != last_type:
            if last_type:
                lines.append("")
            lines.append("// " + "-" * len(type))
            lines.append("// " + type)
            last_type = type
        if last_proj != proj:
            lines.append("")
            lines.append("// " + pathlib.Path(proj).stem)
            last_proj = proj
        if last_lib != lib:
            if lib:
                lines.append("")
            last_lib = lib

        include_directive = "#include <" + include + ">"

        lines.append(include_directive)

    with open(stdafx_h_filename, "w") as f:
        f.write(PREAMBLE)
        f.write("\n".join(lines))
        f.write(POSTAMBLE)


update_stdafx_h("Parsers/StdAfx.h", [], ["Parsers/Parsers.vcxproj"])
update_stdafx_h("Common/StdAfx.h", [], ["Common/Common.vcxproj"])

update_stdafx_h("FreeOrion/StdAfx.h", ["Common/Common.vcxproj", "GiGi/GiGi.vcxproj"], ["FreeOrion/FreeOrion.vcxproj"])
update_stdafx_h("FreeOrionD/StdAfx.h", ["Common/Common.vcxproj"], ["FreeOrionD/FreeOrionD.vcxproj"])
update_stdafx_h("FreeOrionCA/StdAfx.h", ["Common/Common.vcxproj"], ["FreeOrionCA/FreeOrionCA.vcxproj"])

update_stdafx_h("GiGi/StdAfx.h", [], ["GiGi/GiGi.vcxproj"])



from macros.opinion import POLICY_DISLIKE_SCALING
from macros.priorities import AFTER_ALL_TARGET_MAX_METERS_PRIORITY

try:
    from focs._effects import (
        Abs,
        Adequate,
        Capital,
        Contains,
        CurrentTurn,
        EffectsGroup,
        EmpireHasAdoptedPolicy,
        GameRule,
        GasGiantType,
        Good,
        HasSpecies,
        HasTag,
        Homeworld,
        Hostile,
        Huge,
        InSystem,
        IsBuilding,
        IsSource,
        JumpsBetween,
        Large,
        LocalCandidate,
        MaxOf,
        Medium,
        Min,
        MinOf,
        NamedReal,
        OwnedBy,
        Planet,
        Poor,
        Population,
        ResourceSupplyConnected,
        SetHappiness,
        SetTargetHappiness,
        Small,
        Source,
        SpeciesDislikes,
        SpeciesLikes,
        Statistic,
        Target,
        Tiny,
        Uninhabitable,
        Unowned,
        Value,
    )
except ModuleNotFoundError:
    pass


STABILITY_PER_LIKED_FOCUS = 2.0

# Policy Liberty doubles de focus dislike effect, policy Conformance halves it, macro is in macros/opinion.macros
STABILITY_PER_DISLIKED_FOCUS = 2.0 * POLICY_DISLIKE_SCALING

ENVIRONMENT_STABILITY_MODIFIER = [
    EffectsGroup(
        scope=IsSource,
        activation=Planet()
        & Planet(environment=[Uninhabitable])
        & (GameRule(type=int, name="RULE_HOSTILE_ENVIRONMENT_STABILITY") != 0),
        accountinglabel="UNINHABTIABLE_ENVIRONMENT_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_HOSTILE_ENVIRONMENT_STABILITY")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet()
        & Planet(environment=[Hostile])
        & (GameRule(type=int, name="RULE_HOSTILE_ENVIRONMENT_STABILITY") != 0),
        accountinglabel="HOSTILE_ENVIRONMENT_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_HOSTILE_ENVIRONMENT_STABILITY")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet()
        & Planet(environment=[Poor])
        & (GameRule(type=int, name="RULE_POOR_ENVIRONMENT_STABILITY") != 0),
        accountinglabel="POOR_ENVIRONMENT_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_POOR_ENVIRONMENT_STABILITY")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet()
        & Planet(environment=[Adequate])
        & (GameRule(type=int, name="RULE_ADEQUATE_ENVIRONMENT_STABILITY") != 0),
        accountinglabel="ADEQUATE_ENVIRONMENT_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_ADEQUATE_ENVIRONMENT_STABILITY")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet()
        & Planet(environment=[Good])
        & (GameRule(type=int, name="RULE_GOOD_ENVIRONMENT_STABILITY") != 0),
        accountinglabel="GOOD_ENVIRONMENT_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_GOOD_ENVIRONMENT_STABILITY")),
    ),
]

PLANET_SIZE_STABILITY_MODIFIER = [
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Planet(size=[Tiny]) & (GameRule(type=int, name="RULE_TINY_SIZE_STABILITY") != 0),
        accountinglabel="TINY_PLANET_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_TINY_SIZE_STABILITY")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Planet(size=[Small]) & (GameRule(type=int, name="RULE_SMALL_SIZE_STABILITY") != 0),
        accountinglabel="SMALL_PLANET_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_SMALL_SIZE_STABILITY")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Planet(size=[Medium]) & (GameRule(type=int, name="RULE_MEDIUM_SIZE_STABILITY") != 0),
        accountinglabel="SMALL_MEDIUM_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_MEDIUM_SIZE_STABILITY")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Planet(size=[Large]) & (GameRule(type=int, name="RULE_LARGE_SIZE_STABILITY") != 0),
        accountinglabel="LARGE_PLANET_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_LARGE_SIZE_STABILITY")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Planet(size=[Huge]) & (GameRule(type=int, name="RULE_HUGE_SIZE_STABILITY") != 0),
        accountinglabel="HUGE_PLANET_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_HUGE_SIZE_STABILITY")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet()
        & Planet(type=[GasGiantType])
        & (GameRule(type=int, name="RULE_GAS_GIANT_SIZE_STABILITY") != 0),
        accountinglabel="GAS_GIANT_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_GAS_GIANT_SIZE_STABILITY")),
    ),
]


COMMON_HAPPINESS_EFFECTS = [
    *ENVIRONMENT_STABILITY_MODIFIER,
    *PLANET_SIZE_STABILITY_MODIFIER,
    EffectsGroup(  # increase or decrease 1 per turn towards target
        scope=IsSource,
        activation=Planet() & (LocalCandidate.LastTurnConquered < CurrentTurn),
        priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
        effects=SetHappiness(
            value=Value
            + MinOf(float, Abs(float, Value(Target.TargetHappiness) - Value), 1)
            * (1 - 2 * (Value > Value(Target.TargetHappiness)))
        ),
    ),
    EffectsGroup(  # artistic species make other planets with different artistic species on them and in the same system system more stable
        scope=Planet()
        & InSystem(id=Source.SystemID)
        & ~IsSource
        & HasSpecies()
        & ~HasSpecies(name=[Source.Species])
        & HasTag(name="ARTISTIC")
        # (SpeciesSpeciesOpinion species = LocalCandidate.Species species = Source.Species > 0) # TODO: implement within an Or condition, once species-species opinions are working...
        ,
        activation=Planet() & HasSpecies() & HasTag(name="ARTISTIC"),
        accountinglabel="ARTISAN_APPRECIATION",
        effects=SetTargetHappiness(value=Value + 1.0),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=(GameRule(type=int, name="RULE_BASELINE_PLANET_STABILITY") != 0),
        accountinglabel="RULE_BASELINE_PLANET_STABILITY",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_BASELINE_PLANET_STABILITY")),
    ),
    EffectsGroup(  # more stable when liked focuses are adopted
        scope=IsSource,
        activation=Population(low=0.001) & SpeciesLikes(name=LocalCandidate.Focus),
        accountinglabel="LIKES_FOCUS_LABEL",
        effects=SetTargetHappiness(value=Value + STABILITY_PER_LIKED_FOCUS),
    ),
    EffectsGroup(  # less stable when disliked focuses are adopted
        scope=IsSource,
        activation=~Unowned & Population(low=0.001) & SpeciesDislikes(name=LocalCandidate.Focus),
        accountinglabel="DISLIKES_FOCUS_LABEL",
        effects=SetTargetHappiness(value=Value - STABILITY_PER_DISLIKED_FOCUS),
    ),
    EffectsGroup(  # species on their own homeworld are more stable
        scope=IsSource,
        activation=Homeworld(name=[Source.Species]),
        accountinglabel="HOMEWORLD_BONUS",
        effects=SetTargetHappiness(value=Value + 5),
    ),
    EffectsGroup(  # capital planet is more stable
        scope=IsSource,
        activation=Capital,
        accountinglabel="CAPITAL_LABEL",
        effects=SetTargetHappiness(value=Value + 10),
    ),
    EffectsGroup(  # concentration camp nullifies stability
        scope=IsSource,
        activation=Contains(IsBuilding(name=["BLD_CONC_CAMP"])),
        accountinglabel="CONCENTRATION_CAMPS_LABEL",
        priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
        effects=[SetHappiness(value=0), SetTargetHappiness(value=0)],
    ),
]

# this should ensure that there is a capital / admin for JUMPS_TO_CAPITAL_OR_REGAD
SUPPLY_CONNECTED_TO_CAPITAL_OR_REGAD = ResourceSupplyConnected(
    empire=Source.Owner, condition=(Capital | IsBuilding(name=["BLD_REGIONAL_ADMIN"])) & OwnedBy(empire=Source.Owner)
)


def JUMPS_TO_CAPITAL_OR_REGAD(type):
    return Statistic(
        type,
        Min,
        # Replace JumpsBetween by JumpsBetweenByEmpireSupplyConnections once it is implemented
        value=JumpsBetween(Source.ID, LocalCandidate.ID),
        condition=(
            (Capital | IsBuilding(name=["BLD_REGIONAL_ADMIN"]))
            & OwnedBy(empire=Source.Owner)
            & ResourceSupplyConnected(empire=Source.Owner, condition=IsSource)
        ),
    )


STANDARD_SPECIES_CAPITAL_SUPPLY_CONNECTION_STABILITY = [
    EffectsGroup(  # close supply connected to capital planet is more stable
        scope=IsSource,
        activation=Planet()
        & ~Unowned
        & ~Capital
        & SUPPLY_CONNECTED_TO_CAPITAL_OR_REGAD
        & (JUMPS_TO_CAPITAL_OR_REGAD(int) < 5),
        accountinglabel="CAPITAL_CONNECTION_LABEL",
        effects=SetTargetHappiness(value=Value + MinOf(float, 5, 5 - JUMPS_TO_CAPITAL_OR_REGAD(float))),
    ),
    EffectsGroup(  # far supply connected to capital planet is less
        scope=IsSource,
        activation=Planet()
        & ~Unowned
        & ~Capital
        & ~EmpireHasAdoptedPolicy(empire=Source.Owner, name="PLC_CONFEDERATION")
        & SUPPLY_CONNECTED_TO_CAPITAL_OR_REGAD
        & (JUMPS_TO_CAPITAL_OR_REGAD(int) > 5),
        accountinglabel="CAPITAL_POOR_CONNECTION_LABEL",
        effects=SetTargetHappiness(value=Value + MaxOf(float, -5, 5 - JUMPS_TO_CAPITAL_OR_REGAD(float))),
    ),
    EffectsGroup(  # no connection is much less stable
        scope=IsSource,
        activation=Planet()
        & ~Unowned
        & ~Capital
        & ~EmpireHasAdoptedPolicy(empire=Source.Owner, name="PLC_CONFEDERATION")
        & ~SUPPLY_CONNECTED_TO_CAPITAL_OR_REGAD,
        accountinglabel="CAPITAL_DISCONNECTION_LABEL",
        effects=SetTargetHappiness(
            value=Value - NamedReal(name="DISCONNECTED_FROM_CAPITAL_AND_REGIONAL_ADMIN_STABILITY_PENALTY", value=10)
        ),
    ),
]

# BAD_HAPPINESS
# '''
# [[COMMON_HAPPINESS_EFFECTS]]
# [[STANDARD_SPECIES_CAPITAL_SUPPLY_CONNECTION_STABILITY]]

#         EffectsGroup
#             description = "BAD_HAPPINESS_DESC"
#             scope = Source
#             activation = Planet
#             accountinglabel = "BAD_HAPPINESS_LABEL"
#             effects = SetTargetHappiness value = Value + (NamedReal name = "BAD_HAPPINESS_VAL" value = -2.5)
# '''

# AVERAGE_HAPPINESS
# '''
# [[COMMON_HAPPINESS_EFFECTS]]
# [[STANDARD_SPECIES_CAPITAL_SUPPLY_CONNECTION_STABILITY]]
# '''

GOOD_HAPPINESS = [
    *COMMON_HAPPINESS_EFFECTS,
    *STANDARD_SPECIES_CAPITAL_SUPPLY_CONNECTION_STABILITY,
    EffectsGroup(
        description="GOOD_HAPPINESS_DESC",
        scope=IsSource,
        activation=Planet(),
        accountinglabel="GOOD_HAPPINESS_LABEL",
        effects=SetTargetHappiness(value=Value + NamedReal(name="GOOD_HAPPINESS_VAL", value=2.5)),
    ),
]

# GREAT_HAPPINESS
# '''
# [[COMMON_HAPPINESS_EFFECTS]]
# [[STANDARD_SPECIES_CAPITAL_SUPPLY_CONNECTION_STABILITY]]

#         EffectsGroup
#             description = "GREAT_HAPPINESS_DESC"
#             scope = Source
#             activation = Planet
#             accountinglabel = "GREAT_HAPPINESS_LABEL"
#             effects = SetTargetHappiness value = Value + (NamedReal name = "GREAT_HAPPINESS_VAL" value = 5)
# '''

# ULTIMATE_HAPPINESS
# '''
# [[COMMON_HAPPINESS_EFFECTS]]
# [[STANDARD_SPECIES_CAPITAL_SUPPLY_CONNECTION_STABILITY]]

#         EffectsGroup
#             description = "ULTIMATE_HAPPINESS_DESC"
#             scope = Source
#             activation = Planet
#             accountinglabel = "ULTIMATE_HAPPINESS_LABEL"
#             effects = SetTargetHappiness value = Value + (NamedReal name = "ULTIMATE_HAPPINESS_VAL" value = 7.5)
# '''

# INDEPENDENT_HAPPINESS
# '''[[DESCRIPTION_EFFECTSGROUP_MACRO(INDEPENDENT_HAPPINESS_DESC)]]
# [[COMMON_HAPPINESS_EFFECTS]]
# '''


# #include "../macros/opinion.macros"



from logging import error
from typing import IO


def _get_name_from_line(line):
    if "BuildingType " not in line or "BuildingsParser.cpp" not in line or "[trace]" not in line:
        return None
    body = line.rsplit(" : ", 2)[1]
    return body.removeprefix("BuildingType ")


def get_lines_with_dump(f: IO):
    iterator = iter(f)

    for line in iterator:
        if "Start parsing FOCS for BuildingTypes" in line:
            break
    for line in iterator:
        if "End parsing FOCS for BuildingTypes" in line:
            return
        yield line


def parse_buildings(f: IO):
    _result = []
    _current = []
    for line in get_lines_with_dump(f):
        next_section = _get_name_from_line(line)
        if next_section:
            if _current:
                _result.append((_current[0], "".join(_current[1])))
            _current = [next_section, []]
        else:
            if _current:
                _current[1].append(line)

    if not _current:
        error("No data in log, check that game is run with 'parsing' set to 'trace'")
        exit(1)

    _result.append((_current[0], "".join(_current[1])))

    return _result



from macros.misc import DESCRIPTION_EFFECTSGROUP_MACRO

# BROAD_EP and NARROW_EP are stub placeholders for applying the
# board/narrow environment tolerance description to a species.

try:
    from focs._effects import (
        Adequate,
        AsteroidsType,
        Barren,
        Desert,
        GasGiantType,
        Good,
        Hostile,
        Inferno,
        Ocean,
        Poor,
        Radiated,
        Swamp,
        Terran,
        Toxic,
        Tundra,
        Uninhabitable,
    )
except ModuleNotFoundError:
    pass

BROAD_EP = DESCRIPTION_EFFECTSGROUP_MACRO("BROAD_EP_DESC")

# NARROW_EP
# '''[[DESCRIPTION_EFFECTSGROUP_MACRO(NARROW_EP_DESC)]]'''

# SWAMP_STANDARD_EP
# '''environments = [
#       type = Swamp        environment = Good
#       type = Toxic        environment = Adequate
#       type = Inferno      environment = Poor
#       type = Radiated     environment = Hostile
#       type = Barren       environment = Hostile
#       type = Tundra       environment = Hostile
#       type = Desert       environment = Hostile
#       type = Terran       environment = Poor
#       type = Ocean        environment = Adequate
#       type = Asteroids    environment = Uninhabitable
#       type = GasGiant     environment = Uninhabitable
#     ]'''

# TOXIC_STANDARD_EP
# '''environments = [
#         type = Swamp        environment = Adequate
#         type = Toxic        environment = Good
#         type = Inferno      environment = Adequate
#         type = Radiated     environment = Poor
#         type = Barren       environment = Hostile
#         type = Tundra       environment = Hostile
#         type = Desert       environment = Hostile
#         type = Terran       environment = Hostile
#         type = Ocean        environment = Poor
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# INFERNO_STANDARD_EP
# '''environments = [
#         type = Swamp        environment = Poor
#         type = Toxic        environment = Adequate
#         type = Inferno      environment = Good
#         type = Radiated     environment = Adequate
#         type = Barren       environment = Poor
#         type = Tundra       environment = Hostile
#         type = Desert       environment = Hostile
#         type = Terran       environment = Hostile
#         type = Ocean        environment = Hostile
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# RADIATED_STANDARD_EP
# '''environments = [
#         type = Swamp        environment = Hostile
#         type = Toxic        environment = Poor
#         type = Inferno      environment = Adequate
#         type = Radiated     environment = Good
#         type = Barren       environment = Adequate
#         type = Tundra       environment = Poor
#         type = Desert       environment = Hostile
#         type = Terran       environment = Hostile
#         type = Ocean        environment = Hostile
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# BARREN_STANDARD_EP
# '''environments = [
#         type = Swamp        environment = Hostile
#         type = Toxic        environment = Hostile
#         type = Inferno      environment = Poor
#         type = Radiated     environment = Adequate
#         type = Barren       environment = Good
#         type = Tundra       environment = Adequate
#         type = Desert       environment = Poor
#         type = Terran       environment = Hostile
#         type = Ocean        environment = Hostile
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# TUNDRA_STANDARD_EP
# '''environments = [
#         type = Swamp        environment = Hostile
#         type = Toxic        environment = Hostile
#         type = Inferno      environment = Hostile
#         type = Radiated     environment = Poor
#         type = Barren       environment = Adequate
#         type = Tundra       environment = Good
#         type = Desert       environment = Adequate
#         type = Terran       environment = Poor
#         type = Ocean        environment = Hostile
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# DESERT_STANDARD_EP
# '''environments = [
#         type = Swamp        environment = Hostile
#         type = Toxic        environment = Hostile
#         type = Inferno      environment = Hostile
#         type = Radiated     environment = Hostile
#         type = Barren       environment = Poor
#         type = Tundra       environment = Adequate
#         type = Desert       environment = Good
#         type = Terran       environment = Adequate
#         type = Ocean        environment = Poor
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# TERRAN_STANDARD_EP
# '''environments = [
#         type = Swamp        environment = Poor
#         type = Toxic        environment = Hostile
#         type = Inferno      environment = Hostile
#         type = Radiated     environment = Hostile
#         type = Barren       environment = Hostile
#         type = Tundra       environment = Poor
#         type = Desert       environment = Adequate
#         type = Terran       environment = Good
#         type = Ocean        environment = Adequate
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# OCEAN_STANDARD_EP
# '''environments = [
#         type = Swamp        environment = Adequate
#         type = Toxic        environment = Poor
#         type = Inferno      environment = Hostile
#         type = Radiated     environment = Hostile
#         type = Barren       environment = Hostile
#         type = Tundra       environment = Hostile
#         type = Desert       environment = Poor
#         type = Terran       environment = Adequate
#         type = Ocean        environment = Good
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# SWAMP_NARROW_EP
# '''environments = [
#       type = Swamp        environment = Good
#       type = Toxic        environment = Poor
#       type = Inferno      environment = Hostile
#       type = Radiated     environment = Hostile
#       type = Barren       environment = Hostile
#       type = Tundra       environment = Hostile
#       type = Desert       environment = Hostile
#       type = Terran       environment = Hostile
#       type = Ocean        environment = Poor
#       type = Asteroids    environment = Uninhabitable
#       type = GasGiant     environment = Uninhabitable
#     ]'''

# TOXIC_NARROW_EP
# '''environments = [
#         type = Swamp        environment = Poor
#         type = Toxic        environment = Good
#         type = Inferno      environment = Poor
#         type = Radiated     environment = Hostile
#         type = Barren       environment = Hostile
#         type = Tundra       environment = Hostile
#         type = Desert       environment = Hostile
#         type = Terran       environment = Hostile
#         type = Ocean        environment = Hostile
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# INFERNO_NARROW_EP
# '''environments = [
#         type = Swamp        environment = Hostile
#         type = Toxic        environment = Poor
#         type = Inferno      environment = Good
#         type = Radiated     environment = Poor
#         type = Barren       environment = Hostile
#         type = Tundra       environment = Hostile
#         type = Desert       environment = Hostile
#         type = Terran       environment = Hostile
#         type = Ocean        environment = Hostile
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# RADIATED_NARROW_EP
# '''environments = [
#         type = Swamp        environment = Hostile
#         type = Toxic        environment = Hostile
#         type = Inferno      environment = Poor
#         type = Radiated     environment = Good
#         type = Barren       environment = Poor
#         type = Tundra       environment = Hostile
#         type = Desert       environment = Hostile
#         type = Terran       environment = Hostile
#         type = Ocean        environment = Hostile
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# BARREN_NARROW_EP
# '''environments = [
#         type = Swamp        environment = Hostile
#         type = Toxic        environment = Hostile
#         type = Inferno      environment = Hostile
#         type = Radiated     environment = Poor
#         type = Barren       environment = Good
#         type = Tundra       environment = Poor
#         type = Desert       environment = Hostile
#         type = Terran       environment = Hostile
#         type = Ocean        environment = Hostile
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# TUNDRA_NARROW_EP
# '''environments = [
#         type = Swamp        environment = Hostile
#         type = Toxic        environment = Hostile
#         type = Inferno      environment = Hostile
#         type = Radiated     environment = Hostile
#         type = Barren       environment = Poor
#         type = Tundra       environment = Good
#         type = Desert       environment = Poor
#         type = Terran       environment = Hostile
#         type = Ocean        environment = Hostile
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# DESERT_NARROW_EP
# '''environments = [
#         type = Swamp        environment = Hostile
#         type = Toxic        environment = Hostile
#         type = Inferno      environment = Hostile
#         type = Radiated     environment = Hostile
#         type = Barren       environment = Hostile
#         type = Tundra       environment = Poor
#         type = Desert       environment = Good
#         type = Terran       environment = Poor
#         type = Ocean        environment = Hostile
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# TERRAN_NARROW_EP
# '''environments = [
#         type = Swamp        environment = Hostile
#         type = Toxic        environment = Hostile
#         type = Inferno      environment = Hostile
#         type = Radiated     environment = Hostile
#         type = Barren       environment = Hostile
#         type = Tundra       environment = Hostile
#         type = Desert       environment = Poor
#         type = Terran       environment = Good
#         type = Ocean        environment = Poor
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# OCEAN_NARROW_EP
# '''environments = [
#         type = Swamp        environment = Poor
#         type = Toxic        environment = Hostile
#         type = Inferno      environment = Hostile
#         type = Radiated     environment = Hostile
#         type = Barren       environment = Hostile
#         type = Tundra       environment = Hostile
#         type = Desert       environment = Hostile
#         type = Terran       environment = Poor
#         type = Ocean        environment = Good
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# SWAMP_BROAD_EP
# '''environments = [
#       type = Swamp        environment = Good
#       type = Toxic        environment = Adequate
#       type = Inferno      environment = Poor
#       type = Radiated     environment = Poor
#       type = Barren       environment = Hostile
#       type = Tundra       environment = Hostile
#       type = Desert       environment = Poor
#       type = Terran       environment = Poor
#       type = Ocean        environment = Adequate
#       type = Asteroids    environment = Uninhabitable
#       type = GasGiant     environment = Uninhabitable
#     ]'''

# TOXIC_BROAD_EP
# '''environments = [
#         type = Swamp        environment = Adequate
#         type = Toxic        environment = Good
#         type = Inferno      environment = Adequate
#         type = Radiated     environment = Poor
#         type = Barren       environment = Poor
#         type = Tundra       environment = Hostile
#         type = Desert       environment = Hostile
#         type = Terran       environment = Poor
#         type = Ocean        environment = Poor
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

INFERNO_BROAD_EP = {
    Swamp: Poor,
    Toxic: Adequate,
    Inferno: Good,
    Radiated: Adequate,
    Barren: Poor,
    Tundra: Poor,
    Desert: Hostile,
    Terran: Hostile,
    Ocean: Poor,
    AsteroidsType: Uninhabitable,
    GasGiantType: Uninhabitable,
}

# RADIATED_BROAD_EP
# '''environments = [
#         type = Swamp        environment = Poor
#         type = Toxic        environment = Poor
#         type = Inferno      environment = Adequate
#         type = Radiated     environment = Good
#         type = Barren       environment = Adequate
#         type = Tundra       environment = Poor
#         type = Desert       environment = Poor
#         type = Terran       environment = Hostile
#         type = Ocean        environment = Hostile
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# BARREN_BROAD_EP
# '''environments = [
#         type = Swamp        environment = Hostile
#         type = Toxic        environment = Poor
#         type = Inferno      environment = Poor
#         type = Radiated     environment = Adequate
#         type = Barren       environment = Good
#         type = Tundra       environment = Adequate
#         type = Desert       environment = Poor
#         type = Terran       environment = Poor
#         type = Ocean        environment = Hostile
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# TUNDRA_BROAD_EP
# '''environments = [
#         type = Swamp        environment = Hostile
#         type = Toxic        environment = Hostile
#         type = Inferno      environment = Poor
#         type = Radiated     environment = Poor
#         type = Barren       environment = Adequate
#         type = Tundra       environment = Good
#         type = Desert       environment = Adequate
#         type = Terran       environment = Poor
#         type = Ocean        environment = Poor
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# DESERT_BROAD_EP
# '''environments = [
#         type = Swamp        environment = Poor
#         type = Toxic        environment = Hostile
#         type = Inferno      environment = Hostile
#         type = Radiated     environment = Poor
#         type = Barren       environment = Poor
#         type = Tundra       environment = Adequate
#         type = Desert       environment = Good
#         type = Terran       environment = Adequate
#         type = Ocean        environment = Poor
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# TERRAN_BROAD_EP
# '''environments = [
#         type = Swamp        environment = Poor
#         type = Toxic        environment = Poor
#         type = Inferno      environment = Hostile
#         type = Radiated     environment = Hostile
#         type = Barren       environment = Poor
#         type = Tundra       environment = Poor
#         type = Desert       environment = Adequate
#         type = Terran       environment = Good
#         type = Ocean        environment = Adequate
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# OCEAN_BROAD_EP
# '''environments = [
#         type = Swamp        environment = Adequate
#         type = Toxic        environment = Poor
#         type = Inferno      environment = Poor
#         type = Radiated     environment = Hostile
#         type = Barren       environment = Hostile
#         type = Tundra       environment = Poor
#         type = Desert       environment = Poor
#         type = Terran       environment = Adequate
#         type = Ocean        environment = Good
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# TOLERANT_EP
# '''environments = [
#         type = Swamp        environment = Good
#         type = Toxic        environment = Good
#         type = Inferno      environment = Good
#         type = Radiated     environment = Good
#         type = Barren       environment = Good
#         type = Tundra       environment = Good
#         type = Desert       environment = Good
#         type = Terran       environment = Good
#         type = Ocean        environment = Good
#         type = Asteroids    environment = Uninhabitable
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# VERY_TOLERANT_EP
# '''environments = [
#         type = Swamp        environment = Good
#         type = Toxic        environment = Good
#         type = Inferno      environment = Good
#         type = Radiated     environment = Good
#         type = Barren       environment = Good
#         type = Tundra       environment = Good
#         type = Desert       environment = Good
#         type = Terran       environment = Good
#         type = Ocean        environment = Good
#         type = Asteroids    environment = Adequate
#         type = GasGiant     environment = Adequate
#     ]'''

# ASTEROIDAL_NARROW_EP
# '''environments = [
#         type = Swamp        environment = Uninhabitable
#         type = Toxic        environment = Uninhabitable
#         type = Inferno      environment = Uninhabitable
#         type = Radiated     environment = Uninhabitable
#         type = Barren       environment = Hostile
#         type = Tundra       environment = Uninhabitable
#         type = Desert       environment = Uninhabitable
#         type = Terran       environment = Uninhabitable
#         type = Ocean        environment = Uninhabitable
#         type = Asteroids    environment = Adequate
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# ASTEROIDAL_STANDARD_EP
# '''environments = [
#         type = Swamp        environment = Hostile
#         type = Toxic        environment = Hostile
#         type = Inferno      environment = Hostile
#         type = Radiated     environment = Hostile
#         type = Barren       environment = Poor
#         type = Tundra       environment = Hostile
#         type = Desert       environment = Hostile
#         type = Terran       environment = Hostile
#         type = Ocean        environment = Hostile
#         type = Asteroids    environment = Good
#         type = GasGiant     environment = Uninhabitable
#     ]'''

# GASEOUS_NARROW_EP
# '''environments = [
#       type = Swamp        environment = Uninhabitable
#       type = Toxic        environment = Uninhabitable
#       type = Inferno      environment = Uninhabitable
#       type = Radiated     environment = Uninhabitable
#       type = Barren       environment = Uninhabitable
#       type = Tundra       environment = Uninhabitable
#       type = Desert       environment = Uninhabitable
#       type = Terran       environment = Uninhabitable
#       type = Ocean        environment = Hostile
#       type = Asteroids    environment = Uninhabitable
#       type = GasGiant     environment = Adequate
#     ]'''

# GASEOUS_STANDARD_EP
# '''environments = [
#       type = Swamp        environment = Hostile
#       type = Toxic        environment = Hostile
#       type = Inferno      environment = Hostile
#       type = Radiated     environment = Hostile
#       type = Barren       environment = Hostile
#       type = Tundra       environment = Hostile
#       type = Desert       environment = Hostile
#       type = Terran       environment = Hostile
#       type = Ocean        environment = Hostile
#       type = Asteroids    environment = Uninhabitable
#       type = GasGiant     environment = Good
#     ]'''



from macros.base_prod import INDUSTRY_PER_POP
from macros.priorities import (
    TARGET_AFTER_SCALING_PRIORITY,
    TARGET_EARLY_BEFORE_SCALING_PRIORITY,
)

# NO_INDUSTRY
# '''[[DESCRIPTION_EFFECTSGROUP_MACRO(NO_INDUSTRY_DESC)]]'''
try:
    from focs._effects import (
        EffectsGroup,
        Focus,
        GalaxyMaxAIAggression,
        Happiness,
        IsHuman,
        IsSource,
        NamedReal,
        Planet,
        SetTargetIndustry,
        Target,
        TargetIndustry,
        Value,
    )
except ModuleNotFoundError:
    pass


BASIC_INDUSTRY = [
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & TargetIndustry(low=0) & Happiness(low=0) & Focus(type=["FOCUS_INDUSTRY"]),
        accountinglabel="FOCUS_INDUSTRY_LABEL",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        effects=SetTargetIndustry(
            value=Value
            + Target.Population * NamedReal(name="INDUSTRY_FOCUS_TARGET_INDUSTRY_PERPOP", value=1.0 * INDUSTRY_PER_POP)
        ),
    ),
    EffectsGroup(  # gives human bonuses when AI Aggression set to Beginner
        scope=IsSource,
        activation=Planet() & IsHuman & (GalaxyMaxAIAggression == 0),  # human player, not human species
        accountinglabel="DIFFICULTY",
        priority=TARGET_AFTER_SCALING_PRIORITY,
        effects=SetTargetIndustry(value=Value * 2),
    ),
]

# VERY_BAD_INDUSTRY
# '''[[BASIC_INDUSTRY]]

#         EffectsGroup
#             description = "VERY_BAD_INDUSTRY_DESC"
#             scope = Source
#             activation = And [
#                 Planet
#                 TargetIndustry low = 0
#                 Happiness low = 0
#                 Focus type = "FOCUS_INDUSTRY"
#             ]
#             accountinglabel = "VERY_BAD_INDUSTRY_LABEL"
#             priority = [[TARGET_SCALING_PRIORITY]]
#             effects = SetTargetIndustry value = Value
#                         * (NamedReal name = "VERY_BAD_INDUSTRY_TARGET_INDUSTRY_SCALING"
#                                      value = [[VERY_BAD_MULTIPLIER]])
# '''

# BAD_INDUSTRY
# '''[[BASIC_INDUSTRY]]

#         EffectsGroup
#             description = "BAD_INDUSTRY_DESC"
#             scope = Source
#             activation = And [
#                 Planet
#                 TargetIndustry low = 0
#                 Happiness low = 0
#                 Focus type = "FOCUS_INDUSTRY"
#             ]
#             accountinglabel = "BAD_INDUSTRY_LABEL"
#             priority = [[TARGET_SCALING_PRIORITY]]
#             effects = SetTargetIndustry value = Value
#                         * (NamedReal name = "BAD_INDUSTRY_TARGET_INDUSTRY_SCALING"
#                                      value = [[BAD_MULTIPLIER]])
# '''

AVERAGE_INDUSTRY = BASIC_INDUSTRY

# GOOD_INDUSTRY
# '''[[BASIC_INDUSTRY]]

#         EffectsGroup
#             description = "GOOD_INDUSTRY_DESC"
#             scope = Source
#             activation = And [
#                 Planet
#                 TargetIndustry low = 0
#                 Happiness low = 0
#                 Focus type = "FOCUS_INDUSTRY"
#             ]
#             accountinglabel = "GOOD_INDUSTRY_LABEL"
#             priority = [[TARGET_SCALING_PRIORITY]]
#             effects = SetTargetIndustry value = Value
#                         * (NamedReal name = "GOOD_INDUSTRY_TARGET_INDUSTRY_SCALING"
#                                      value = [[GOOD_MULTIPLIER]])
# '''

# GREAT_INDUSTRY
# '''[[BASIC_INDUSTRY]]

#         EffectsGroup
#             description = "GREAT_INDUSTRY_DESC"
#             scope = Source
#             activation = And [
#                 Planet
#                 TargetIndustry low = 0
#                 Happiness low = 0
#                 Focus type = "FOCUS_INDUSTRY"
#             ]
#             accountinglabel = "GREAT_INDUSTRY_LABEL"
#             priority = [[TARGET_SCALING_PRIORITY]]
#             effects = SetTargetIndustry value = Value
#                         * (NamedReal name = "GREAT_INDUSTRY_TARGET_INDUSTRY_SCALING"
#                                      value = [[GREAT_MULTIPLIER]])
# '''

# ULTIMATE_INDUSTRY
# '''[[BASIC_INDUSTRY]]

#         EffectsGroup
#             description = "ULTIMATE_INDUSTRY_DESC"
#             scope = Source
#             activation = And [
#                 Planet
#                 TargetIndustry low = 0
#                 Happiness low = 0
#                 Focus type = "FOCUS_INDUSTRY"
#             ]
#             accountinglabel = "ULTIMATE_INDUSTRY_LABEL"
#             priority = [[TARGET_SCALING_PRIORITY]]
#             effects = SetTargetIndustry value = Value
#                         * (NamedReal name = "ULTIMATE_INDUSTRY_TARGET_INDUSTRY_SCALING"
#                                      value = [[ULTIMATE_MULTIPLIER]])
# '''



# Policy Liberty doubles the stability malus from disliked buildings, policies and planetary foci
# Policy Conformance halves it
# Values are declared in macros/named_values.focs.txt
try:
    from focs._effects import EmpireHasAdoptedPolicy, IsSource, StatisticIf, Target
except ModuleNotFoundError:
    pass


POLICY_DISLIKE_SCALING = (
    1.0
    * (
        2.0
        ** (StatisticIf(float, condition=IsSource & EmpireHasAdoptedPolicy(empire=Target.Owner, name="PLC_LIBERTY")))
    )
    * (
        0.5
        ** (
            StatisticIf(float, condition=IsSource & EmpireHasAdoptedPolicy(empire=Target.Owner, name="PLC_CONFORMANCE"))
        )
    )
)



from pathlib import Path
from subprocess import getoutput


def get_branch():
    return getoutput(["git", "rev-parse", "--abbrev-ref", "HEAD"]).replace("/", "-")


def get_base_dump_folder():
    return Path(__file__).parent / "dumps"



from io import StringIO

from check.focs_dump_parser.parser import _get_name_from_line, parse_buildings

file_ = StringIO(
    """
00:00:34.002036 {0x00002980} [info] log : Logger.cpp:137 : Added logger named "conditions"
00:00:34.001036 {0x00002980} [trace] parsing : BuildingsParser.cpp:258 : Start parsing FOCS for BuildingTypes: 108
00:00:34.002036 {0x00002980} [debug] log : LoggerWithOptionsDB.cpp:108 : Configure log source "effects" from optionsDB using threshold debug
00:00:34.002036 {0x00002980} [trace] parsing : BuildingsParser.cpp:260 : BuildingType BLD_ABANDON_OUTPOST : 4243737
BuildingType
    name = "BLD_ABANDON_OUTPOST"
    description = "BLD_ABANDON_OUTPOST_DESC"
00:00:34.002036 {0x00002980} [trace] parsing : BuildingsParser.cpp:260 : BuildingType BLD_ART_BLACK_HOLE : 4128519
BuildingType
    name = "BLD_ART_BLACK_HOLE"
    description = "BLD_ART_BLACK_HOLE_DESC"
00:00:34.029042 {0x00002980} [trace] parsing : BuildingsParser.cpp:261 : End parsing FOCS for BuildingTypes108
""".strip()
)


bld_abandon_outpost = """
BuildingType
    name = "BLD_ABANDON_OUTPOST"
    description = "BLD_ABANDON_OUTPOST_DESC"
""".lstrip()

bld_art_black_hole = """
BuildingType
    name = "BLD_ART_BLACK_HOLE"
    description = "BLD_ART_BLACK_HOLE_DESC"
""".lstrip()


def test_parse():
    assert parse_buildings(file_) == [
        ("BLD_ABANDON_OUTPOST", bld_abandon_outpost),
        ("BLD_ART_BLACK_HOLE", bld_art_black_hole),
    ]


def test_get_name_from_line():
    assert (
        _get_name_from_line(
            "00:00:34.002036 {0x00002980} [trace] parsing : BuildingsParser.cpp:260 : BuildingType BLD_ABANDON_OUTPOST : 4243737"
        )
        == "BLD_ABANDON_OUTPOST"
    )
    assert (
        _get_name_from_line(
            '00:00:34.002036 {0x00002980} [info] log : Logger.cpp:257 : Setting "conditions" logger threshold to "debug".'
        )
        is None
    )



# HUGE_PLANET
# '''EffectsGroup
#         scope = Source
#         activation = And [
#             Planet
#             Turn high = 0
#         ]
#         effects = SetPlanetSize planetsize = Huge
# '''
try:
    from focs._effects import EffectsGroup, IsSource, Large, Planet, SetPlanetSize, Turn
except ModuleNotFoundError:
    pass


LARGE_PLANET = [
    EffectsGroup(scope=IsSource, activation=Planet() & Turn(high=0), effects=SetPlanetSize(planetsize=Large))
]

# MEDIUM_PLANET
# '''EffectsGroup
#         scope = Source
#         activation = And [
#             Planet
#             Turn high = 0
#         ]
#         effects = SetPlanetSize planetsize = Medium
# '''

# SMALL_PLANET
# '''EffectsGroup
#         scope = Source
#         activation = And [
#             Planet
#             Turn high = 0
#         ]
#         effects = SetPlanetSize planetsize = Small
# '''

# NOT_HUGE_PLANET
# '''EffectsGroup
#         scope = Source
#         activation = And [
#             Planet
#             Turn high = 0
#             Planet size = Huge
#         ]
#         effects = SetPlanetSize planetsize = Large
# '''

# NOT_LARGE_PLANET
# '''EffectsGroup
#         scope = Source
#         activation = And [
#             Planet
#             Turn high = 0
#             Not Planet size = Tiny
#             Not Planet size = Medium
#         ]
#         effects = SetPlanetSize planetsize = Small
# '''



#!/usr/bin/python3
import argparse
import datetime
import html
import json
import os
import re
import subprocess
import sys
import textwrap
import threading
import time
import urllib.parse
import urllib.request
import webbrowser
from collections import OrderedDict
from collections.abc import Iterator
from http.server import BaseHTTPRequestHandler, HTTPServer
from typing import NamedTuple

STRING_TABLE_KEY_PATTERN = re.compile(r"^[A-Z0-9_]+$")
# Provides the named capture groups 'ref_type' and 'key'
INTERNAL_REFERENCE_PATTERN_TEMPLATE = r"\[\[(?P<ref_type>(?:(?:encyclopedia|(?:building|field|meter)type|predefinedshipdesign|ship(?:hull|part)|special|species|tech|policy|value|browsepath) )?)(?P<key>{})\]\]"
OPTIONAL_REF_TYPES = ["value", "browsepath"]


def _get_brackets(text) -> Iterator[str]:
    has = False
    prev = ""
    for char in text:
        if char in "[]":
            if has:
                has = False
                if prev == char:
                    yield char
            else:
                has = True
                prev = char
        else:
            has = False


def check_balanced_reference(text: str) -> bool:
    """
    Check if reference [[XXX]] has balanced syntax.

    >>> check_balanced_reference("no reference")
    True

    >>> check_balanced_reference("[[valid reference]]")
    True

    >>> check_balanced_reference("[[unbalanced left]")
    False

    >>> check_balanced_reference("[unbalanced right]]")
    False

    >>> check_balanced_reference("[[ multiple with ]] [[ not closed")
    False

    >>> check_balanced_reference("[[ [[ reference ]] in reference ]]")
    False
    """
    braces = _get_brackets(text)
    is_open = False
    for b in braces:
        if b == "[":
            if is_open:
                return False
            is_open = True
        else:
            if not is_open:
                return False
            is_open = False
    return not is_open


class StringTableEntry:
    def __init__(self, key, keyline, value, notes, value_times):
        if not STRING_TABLE_KEY_PATTERN.match(key):
            raise ValueError(f"Key doesn't match expected format [A-Z0-9_]+, was '{key}'")
        self.key = key
        self.keyline = keyline
        self.value = value
        self.value_times = value_times
        self.notes = notes

    def __str__(self):
        result = ""
        for note in self.notes:
            result += "# " + note + "\n"
        if self.value is not None:
            result += self.key + "\n"
            if not self.value or "\n" in self.value or re.search(r"^\s", self.value) or re.search(r"\s$", self.value):
                result += "'''" + self.value + "'''" + "\n"
            else:
                result += self.value + "\n"
        else:
            result += "#*" + self.key + "\n"
            for e in self.value_times:
                result += f"#* <not translated {datetime.datetime.fromtimestamp(e).isoformat(sep=' ')}>\n"

        return result

    def __repr__(self):
        return f"StringTableEntry(key={self.key}, keyline={self.keyline}, value={self.value})"


class StringTable:
    def __init__(self, fpath, language, notes, includes, entries):
        self.fpath = fpath
        self.language = language
        self.notes = notes
        self.includes = includes
        self._entries = entries
        self._ientries = OrderedDict()

        for entry in self._entries:
            if isinstance(entry, StringTableEntry):
                if entry.key in self._ientries:
                    raise ValueError(
                        f"{self.fpath}:{entry.keyline}:"
                        f" Duplicate key '{entry.key}', previously defined on line {self._ientries[entry.key].keyline}"
                    )
                self._ientries[entry.key] = entry

    def __getitem__(self, key):
        return self._ientries[key]

    def __contains__(self, key):
        return key in self._ientries

    def __iter__(self):
        return iter(self._ientries)

    def __len__(self):
        return len(self._ientries)

    def __str__(self):  # noqa: C901
        result = ""
        result += self.language + "\n"

        if self.notes:
            result += "\n"
        for note in self.notes:
            if note:
                result += "# " + note + "\n"
            else:
                result += "#\n"

        for entry in self._entries:
            if isinstance(entry, StringTableEntry):
                result += "\n"
                result += str(entry)
            elif isinstance(entry, list):
                result += "\n"
                result += "\n"
                result += "##\n"
                for line in entry:
                    if line:
                        result += "## " + line + "\n"
                    else:
                        result += "##\n"
                result += "##\n"

        if self.includes:
            result += "\n"

        for include in self.includes:
            result += '#include "' + include + '"\n'

        return result

    def __repr__(self):
        return f"StringTable(fpath={self.fpath}, language={self.language}, entries={self.entries})"

    def items(self):
        return self._entries

    @staticmethod
    def set_author(fpath, entries, blames):  # noqa: C901
        blame_cmd = ["git", "blame", "--incremental", fpath]
        git_blame = subprocess.check_output(blame_cmd)
        git_blame = git_blame.decode("utf-8", "strict")
        git_blame = git_blame.splitlines(keepends=True)

        value_times = {}
        author_time = None
        line = None
        line_range = None

        # Provides the named capture groups 'sha', 'old_line', 'new_line' and 'range'
        GIT_BLAME_INCREMENTAL_PATTERN = re.compile(
            r"^(?P<sha>[0-9a-f]{40}) (?P<old_line>[1-9][0-9]*) (?P<new_line>[1-9][0-9]*)(?P<range> [1-9][0-9]*)?$"
        )

        for line_blame in git_blame:
            match = GIT_BLAME_INCREMENTAL_PATTERN.match(line_blame)
            if match:
                line = int(match["new_line"])
                line_range = int(match["range"])
            if line_blame.startswith("author-time "):
                author_time = int(line_blame.strip().split(" ")[1])
            if line_blame.startswith("filename "):
                for line in range(line, line + line_range):
                    if line not in blames:
                        continue
                    line_times = value_times.get(blames[line], [])
                    line_times.append(author_time)
                    value_times[blames[line]] = line_times

        for entry in entries:
            if isinstance(entry, StringTableEntry):
                if not entry.value:
                    # ignore untranslated entries
                    continue
                if entry.key not in value_times or len(value_times[entry.key]) != len(entry.value_times):
                    raise RuntimeError(
                        f"{fpath}: git blame did not collect any matching author times for key {entry.key}"
                    )
                entry.value_times = value_times[entry.key]

    @staticmethod
    def from_file(fhandle, with_blame=True) -> "StringTable":
        return StringTable.from_text(fhandle.read(), fhandle.name, with_blame=with_blame)

    @staticmethod
    def from_text(text, fpath, with_blame=True) -> "StringTable":  # noqa: C901
        is_quoted = False

        language = None
        fnotes = []
        section = []
        entries = []
        includes = []
        blames = OrderedDict()

        key = None
        keyline = None
        value = None
        vline_start = 0
        vline_end = 0
        notes = []
        untranslated_key = None
        untranslated_keyline = None
        untranslated_lines = []

        for fline, line in enumerate(text.splitlines(keepends=True), 1):
            if not is_quoted and not line.strip():
                # discard empty lines when not in quoted value
                if section:
                    while not section[0]:
                        section.pop(0)
                    while not section[-1]:
                        section.pop(-1)
                    entries.append(section)
                    section = []
                if language and notes and not entries:
                    fnotes = notes
                    notes = []
                if untranslated_key:
                    try:
                        entries.append(
                            StringTableEntry(untranslated_key, untranslated_keyline, None, notes, untranslated_lines)
                        )
                    except ValueError as e:
                        raise ValueError(f"{fpath}:{keyline}: {str(e)}")
                    untranslated_key = None
                    untranslated_keyline = None
                    notes = []
                    untranslated_lines = []
                vline_start = vline_end = fline
                continue

            if not is_quoted and line.startswith("#"):
                # discard comments when not in quoted value
                if line.startswith("#include"):
                    includes.append(line[8:].strip()[1:-1].strip())
                if line.startswith("##"):
                    section.append(line[3:].rstrip())
                elif line.startswith("#*") and not untranslated_key:
                    untranslated_key = line[2:].strip()
                    untranslated_keyline = fline
                elif line.startswith("#*"):
                    untranslated_lines.append(0)
                elif line.startswith("#"):
                    notes.append(line[2:].rstrip())
                vline_start = vline_end = fline
                continue

            if not language:
                # read first non comment and non empty line as language name
                language = line.strip()
                vline_start = vline_end = fline
                continue

            if not key:
                if section:
                    while not section[0]:
                        section.pop(0)
                    while not section[-1]:
                        section.pop(-1)
                    entries.append(section)
                    section = []
                if language and notes and not entries:
                    fnotes = notes
                    notes = []
                if untranslated_key:
                    try:
                        entries.append(
                            StringTableEntry(untranslated_key, untranslated_keyline, None, notes, untranslated_lines)
                        )
                    except ValueError as e:
                        raise ValueError(f"{fpath}:{keyline}: {str(e)}")
                    untranslated_key = None
                    untranslated_keyline = None
                    notes = []
                    untranslated_lines = []
                # read non comment and non empty line after language and after
                # completing an stringtable language
                key = line.strip().split()[0]
                keyline = fline
                vline_start = vline_end = fline
                continue

            if not is_quoted:
                if line.startswith("'''"):
                    # prepare begin of quoted value
                    line = line[3:]
                    is_quoted = True
                else:
                    # prepare unquoted value
                    line = line.strip()
                    vline_end = fline
                vline_start = fline

            if is_quoted and line.rstrip().endswith("'''"):
                # prepare end of quoted value
                line = line.rstrip()[:-3]
                is_quoted = False
                vline_end = fline

            # extract value or concatenate quoted value
            value = (value if value else "") + line

            if not is_quoted and key is not None and value is not None:
                # consume collected string table entry
                try:
                    value_range = {line: key for line in range(vline_start, vline_end + 1)}
                    entries.append(StringTableEntry(key, keyline, value, notes, [0] * len(value_range)))
                    blames.update(value_range)
                except ValueError as e:
                    raise ValueError(f"{fpath}:{keyline}: {e}")

                key = None
                keyline = None
                value = None
                notes = []
                vline_start = vline_end = fline
                continue

            if is_quoted:
                # continue consuming quoted value lines
                vline_end = fline
                continue

            raise ValueError(f"{fpath}:{vline_start}: Impossible parser state; last valid line: {line}")

        if key:
            raise ValueError(f"{fpath}:{vline_start}: Key '{key}' without value")

        if is_quoted:
            raise ValueError(f"{fpath}:{vline_start}: Quotes not closed for key '{key}'")

        if with_blame:
            StringTable.set_author(fpath, entries, blames)
        return StringTable(fpath, language, fnotes, includes, entries)

    @staticmethod
    def statistic(left: "StringTable", right: "StringTable"):  # noqa: C901
        class STStatistic(NamedTuple):
            left: "StringTable"
            left_only: set
            left_older: set
            left_pure_reference: set
            right: "StringTable"
            right_only: set
            right_older: set
            right_pure_reference: set
            untranslated: set
            identical: set
            layout_mismatch: set

        all_keys = set(left).union(right)

        untranslated = set()
        identical = set()
        layout_mismatch = set()
        left_only = set()
        right_only = set()
        left_older = set()
        right_older = set()
        left_pure_reference = set()
        right_pure_reference = set()

        for key in all_keys:
            if key in left and key not in right:
                left_only.add(key)
            if key not in left and key in right:
                right_only.add(key)
            if key in left and key in right:
                if not left[key].value or not right[key].value:
                    untranslated.add(key)
                    continue
                if left[key].value == right[key].value:
                    identical.add(key)
                if (
                    left[key].value.startswith("[[")
                    and 2 == left[key].value.count("[")
                    and left[key].value.endswith("]]")
                    and 2 == left[key].value.count("]")
                ):
                    left_pure_reference.add(key)
                if (
                    right[key].value.startswith("[[")
                    and 2 == right[key].value.count("[")
                    and right[key].value.endswith("]]")
                    and 2 == right[key].value.count("]")
                ):
                    right_pure_reference.add(key)
                if len(left[key].value_times) != len(right[key].value_times):
                    layout_mismatch.add(key)
                    continue
                value_times = list(zip(left[key].value_times, right[key].value_times))
                if any([e[0] > e[1] for e in value_times]):
                    right_older.add(key)
                if any([e[0] < e[1] for e in value_times]):
                    left_older.add(key)

        return STStatistic(
            left=left,
            left_only=left_only,
            left_older=left_older,
            right=right,
            right_only=right_only,
            right_older=right_older,
            left_pure_reference=left_pure_reference,
            right_pure_reference=right_pure_reference,
            untranslated=untranslated,
            identical=identical,
            layout_mismatch=layout_mismatch,
        )

    @staticmethod
    def sync(reference, source):
        entries = []

        for item in reference.items():
            if isinstance(item, StringTableEntry):
                if item.key in source:
                    source_item = source[item.key]
                    entries.append(
                        StringTableEntry(
                            item.key,
                            item.keyline,
                            source_item.value if source_item.value != item.value else None,
                            item.notes,
                            source_item.value_times,
                        )
                    )
                else:
                    entries.append(
                        StringTableEntry(item.key, item.keyline, None, item.notes, [0] * len(item.value_times))
                    )
            else:
                entries.append(item)

        return StringTable(source.fpath, source.language, source.notes, source.includes, entries)


class EditServerHandler(BaseHTTPRequestHandler):
    body = """
<html>
    <head>
        <title>Editing {source}</title>
        <style>
            body {{
                font-family: sans-serif;
            }}
            .menu {{
                border: None;
                margin: 0;
                padding: 0;
                position: relative;
            }}
            .menu legend {{
                cursor: pointer;
                font-weight: bold;
                margin-bottom: 0.2em;
            }}
            #display-settings {{
                display: inline;
            }}
            #display-settings > label {{
                display: none;
            }}
            table {{
                table-layout: fixed;
                width: 100%;
            }}
            td {{
                vertical-align: top;
                white-space: pre-wrap;
            }}
            td.source, td.translation {{
                border-style: solid;
                border-width: thin;
                padding-left: 0.5em;
                padding-right: 0.5em;
            }}
            td.source {{
                margin-right: 2em;
            }}
            td.translation {{
                position: relative;
                padding-left: 1em;
                border-left-width: thick;
            }}
            td.translation:before {{
                position:  absolute;
                color: #fff;
                border-radius:  100%;
                width: 1.2em;
                height: 1.2em;
                text-align: center !important;
                vertical-algin: middle;
                top: 0.1em;
                left: -0.8em;
                text-align: center;
                font-weight: bolder;
                line-height: 1.2em;
            }}
            .status-recent td.translation {{
                border-color: #00cc00;
                background-color: #00cc0010;
                color: black;
            }}
            .status-recent td.translation:before {{
                content: "\\2713";
                background-color: #00cc00;
            }}
            .status-stale td.translation {{
                border-color: #cc8400;
                background-color: #cc840010;
                color: black;
            }}
            .status-stale td.translation:before {{
                content: "!";
                background-color: #cc8400;
            }}
            .status-untranslated td.translation {{
                border-color: #ff0000;
                background-color: #ff000010;
                color: black;
            }}
            .status-untranslated td.translation:before {{
                content: "\\274c";
                background-color: #ff0000;
            }}
            .entry td.source {{
                border-color: #808080;
                background-color: #80808010;
                color: black;
            }}
            form.translator {{
                width: 100%;
            }}
        </style>
        <script type="text/javascript" src="/jquery.js"></script>
        <script type="text/javascript" src="/jeditable.jquery.js"></script>
        <script type="text/javascript">
            function toggle_display_settings() {{
                visible = ('true' == this.dataset.visible) || false;

                if(visible) {{
                    document.getElementById('toggle-display-indicator').innerHTML = '&#x25b6;';
                    Array.from(
                        this.parentElement.getElementsByTagName('label')
                    ).forEach(function(element, idx) {{
                        element.style.display = 'none';
                    }});
                }} else {{
                    document.getElementById('toggle-display-indicator').innerHTML = '&#x25bd;';
                    Array.from(
                        this.parentElement.getElementsByTagName('label')
                    ).forEach(function(element, idx) {{
                        element.style.display = 'block';
                    }});
                }}

                this.dataset.visible = (!visible).toString();
            }}

            function toggle_entries_by_status(status) {{
                checkbox = document.getElementById('toggle-' + status);
                checkbox.disabled = true;
                visible = ('true' == checkbox.dataset.visible) || false;

                if(visible) {{
                    setTimeout(function() {{
                        Array.from(
                            document.getElementsByClassName('status-' + status)
                        ).forEach(function(element, idx) {{
                            element.style.display = '';
                        }});
                        document.getElementById('toggle-' + status).disabled = false;
                    }});
                }} else {{
                    setTimeout(function() {{
                        Array.from(
                            document.getElementsByClassName('status-' + status)
                        ).forEach(function(element, idx) {{
                            element.style.display = 'none';
                        }});
                        document.getElementById('toggle-' + status).disabled = false;
                    }});
                }}

                checkbox.dataset.visible = (!visible).toString();
            }}

            function save_entry(value, settings, response_handler) {{
                $.post({{
                    url: '/update',
                    data: JSON.stringify({{"id": settings.id, "value": value}}),
                    contenttype: 'application/json; charset=UTF-8',
                    datatype: 'json',
                    success: function(response) {{
                        var el = document.getElementById(response.id);
                        el.classList.remove('status-recent', 'status-stale', 'status-untranslated');
                        el.classList.add('status-'+response.status);
                        response_handler($('<div />').text(value).html(), true);
                    }}
                }});
            }}

            window.addEventListener('unload', function() {{
                navigator.sendBeacon('/shutdown');
            }});

            document.addEventListener('DOMContentLoaded', function() {{
                var el = document.getElementById('toggle-display-settings');
                el.onclick = toggle_display_settings.bind(el);
                var el = document.getElementById('toggle-recent');
                el.onclick = toggle_entries_by_status.bind(el, 'recent');
                var el = document.getElementById('toggle-stale');
                el.onclick = toggle_entries_by_status.bind(el, 'stale');
                var el = document.getElementById('toggle-untranslated');
                el.onclick = toggle_entries_by_status.bind(el, 'untranslated');

                Array.from(document.getElementsByClassName('translation')).forEach(function(element, idx) {{
                    $(element).editable(save_entry, {{
                        type: 'textarea',
                        id: element.dataset.id,
                        submit: 'Save',
                        cancel: 'Cancel',
                        placeholder: '<i>untranslated</i>',
                        cssclass: 'translator',
                        width: '100%',
                        rows: 5
                    }});
                }});
            }});
        </script>
    </head>
    <body>
        <h1>Translate stringtable for language &quot;{language}&quot;</h1>
        <form action="save" method="post">
            <fieldset class="menu">
                <legend>Stringtable</legend>
                <button type="submit" id="save-stringtable">Save</button>
            </fieldset>
        </form>
        <fieldset class="menu" id="display-settings">
            <legend id="toggle-display-settings"><span id="toggle-display-indicator">&#x25b6;</span> Display</legend>
            <label><input class="input-slider" type="checkbox" checked="checked" id="toggle-recent">Recent translations</label>
            <label><input class="input-slider" type="checkbox" checked="checked" id="toggle-stale">Stale translations</label>
            <label><input class="input-slider" type="checkbox" checked="checked" id="toggle-untranslated">Untranslated translations</label>
        </fieldset>
        <table>
            {entries}
        </table>
    </body>
</html>
    """

    entry = """
<tbody class="entry status-{status}" id="{key}">
    <tr><th colspan="2">{key}</th></tr>
    <tr class="notes"><td colspan="2">{notes}</td></tr>
    <tr><td class="source">{source}</td><td class="translation" data-id="{key}">{translation}</td></tr>
</tbody>
    """

    def do_GET(self):
        parsed_path = urllib.parse.urlparse(self.path)

        if parsed_path.path == "/":
            return self.GET_stringtable()
        elif parsed_path.path == "/jquery.js":
            return self.GET_url("https://code.jquery.com/jquery-3.4.1.min.js")
        elif parsed_path.path == "/jeditable.jquery.js":
            return self.GET_url(
                "https://raw.githubusercontent.com/NicolasCARPi/jquery_jeditable/2.0.14/dist/jquery.jeditable.min.js"
            )
        else:
            self.send_response(404)
            self.end_headers()

    def do_POST(self):
        parsed_path = urllib.parse.urlparse(self.path)

        if parsed_path.path == "/update":
            return self.POST_entry()
        if parsed_path.path == "/save":
            return self.POST_save()
        if parsed_path.path == "/shutdown":
            # The http.server.HttpServer states, that the shutdown function
            # needs to be called outside of the thread, which processes the
            # handler, to avoid deadlock.
            threading.Thread(target=(lambda: self.server.shutdown())).start()
        else:
            self.send_response(404)
            self.end_headers()

    def GET_url(self, url):
        response = urllib.request.urlopen(url)
        self.send_response(response.getcode())
        self.send_header("Content-Type", response.info()["Content-Type"])
        self.end_headers()

        for chunk in response:
            self.wfile.write(chunk)

    def GET_stringtable(self):
        self.send_response(200)
        self.send_header("Content-Type", "text/html ; charset=utf-8")
        self.end_headers()

        entries = []

        stat = StringTable.statistic(self.server.reference_st, self.server.source_st)

        for key in self.server.reference_st:
            source = self.server.reference_st[key].value
            source = html.escape(source)

            translation = self.server.source_st[key].value if key in self.server.source_st else ""

            status = "recent"
            if key not in stat.left_older:
                status = "stale"
            if not translation:
                status = "untranslated"

            translation = translation if translation else ""
            translation = html.escape(translation)

            notes = self.server.reference_st[key].notes
            notes = "\n".join(notes)
            notes = html.escape(notes)

            entries.append(
                EditServerHandler.entry.format(
                    key=key, source=source, translation=translation, notes=notes, status=status
                )
            )

        self.wfile.write(
            EditServerHandler.body.format(
                source=self.server.source_st.fpath, language=self.server.source_st.language, entries="".join(entries)
            ).encode("utf-8")
        )

    def POST_error(self, message):
        self.send_response(500)
        self.send_header("Content-Type", "application/json")
        self.end_headers()

        self.wfile.write(json.dumps({"code": 500, "message": message}).encode("utf-8"))

    def POST_save(self):
        self.send_response(200)
        self.send_header("Content-Type", "text/plain ; charset=utf-8")
        self.send_header(
            "Content-Disposition", f'attachment; filename="{os.path.basename(self.server.source_st.fpath)}"'
        )
        self.end_headers()

        self.wfile.write(str(self.server.source_st).encode("utf-8"))

    def POST_entry(self):
        try:
            content_length = int(self.headers.get("Content-Length", "0"))
            body = self.rfile.read(content_length)
            request = json.loads(body)
            if "id" not in request or "value" not in request:
                raise ValueError("Some of the expected json keys are missing")
        except ValueError:
            self.POST_error("Browser didn't send a save request")
            raise

        if request["id"] not in self.server.reference_st:
            self.POST_error(f"No entry with key {request['id']}")

        entry = self.server.source_st[request["id"]]
        rentry = self.server.reference_st[request["id"]]

        if entry.value != request["value"]:
            if request["value"] and rentry.value != request["value"]:
                entry.value = request["value"]
            else:
                entry.value = None
            if entry.value:
                entry.value_times = [int(time.time())] * len(entry.value.split("\n"))
            else:
                entry.value_times = rentry.value_times

        self.send_response(200)
        self.send_header("Content-Type", "application/json")
        self.end_headers()

        status = "recent"
        if max(self.server.reference_st[request["id"]].value_times) > min(entry.value_times):
            status = "stale"
        if not entry.value:
            status = "untranslated"

        response = {
            "id": request["id"],
            "status": status,
        }

        self.wfile.write(json.dumps(response).encode("utf-8"))


def format_action(args):
    """
    Format source file inline and exit.

    If file was formatted exit status will be 1.
    """
    path = args.source.name
    text = args.source.read()
    source_st = StringTable.from_text(text=text, fpath=path)
    new_text = str(source_st)

    if text != new_text:
        with open(path, "w") as f:
            f.write(new_text)
        print(f"{path} was reformatted")
        return 1
    else:
        print(f"{path} is already well formatted")
        return 0


def sync_action(args):
    reference_st = StringTable.from_file(args.reference)
    source_st = StringTable.from_file(args.source)

    out_st = StringTable.sync(reference_st, source_st)

    print(out_st, end="")

    return 0


def rename_key_action(args):
    if not STRING_TABLE_KEY_PATTERN.match(args.old_key):
        raise ValueError(f"The given old key '{args.old_key}' is not a valid name")

    if not STRING_TABLE_KEY_PATTERN.match(args.new_key):
        raise ValueError(f"The given new key '{args.new_key}' is not a valid name")

    source_st = StringTable.from_file(args.source)

    for key in source_st:
        entry = source_st[key]
        entry.value = re.sub(
            INTERNAL_REFERENCE_PATTERN_TEMPLATE.format(re.escape(args.old_key)),
            r"[[\1" + args.new_key + "]]",
            entry.value,
        )
        if entry.key == args.old_key:
            entry.key = args.new_key

    print(source_st, end="")

    return 0


def edit_action(args):
    reference_st = StringTable.from_file(args.reference)
    source_st = StringTable.from_file(args.source)

    source_st = StringTable.sync(reference_st, source_st)

    server = HTTPServer(("localhost", 8080), EditServerHandler)
    server.reference_st = reference_st
    server.source_st = source_st
    print("Starting server at http://localhost:8080/")
    webbrowser.open("http://localhost:8080/")
    try:
        server.serve_forever()
    except (KeyboardInterrupt, EOFError):
        print("Stopping server")

    return 0


def check_action(args):
    exit_code = 0
    reference_st = None
    if args.reference:
        reference_st = StringTable.from_file(args.reference, with_blame=False)
    for source in args.sources:
        source_st = StringTable.from_file(source, with_blame=False)

        references = set()
        for key in source_st:
            entry = source_st[key]

            if entry.value:
                if not check_balanced_reference(entry.value):
                    print(f"{source_st.fpath}:{entry.key}: Unbalanced brackets: '{entry.value}'")
                    exit_code = 1

                for match in re.finditer(INTERNAL_REFERENCE_PATTERN_TEMPLATE.format(".*?"), entry.value):
                    reference_key = match["key"]
                    references.add(reference_key)
                    if reference_key not in source_st:
                        if match["ref_type"].strip() in OPTIONAL_REF_TYPES:
                            continue

                        print(
                            f"{source_st.fpath}:{entry.keyline}: Referenced key '{reference_key}' in value of '{entry.key}' was not found."
                        )
                        exit_code = 1

        exit_code += _check_key_usage(source_st, reference_st, references)
    return exit_code


def _check_key_usage(source_st, reference_st, references):
    exit_code = 0

    def check_key_is_used(key_under_check):
        if not reference_st:
            return True
        if key_under_check in reference_st:
            return True
        if key_under_check in references:
            return True
        return False

    for key in source_st:
        if not check_key_is_used(key):
            print(f"{source_st.fpath}:{source_st[key].keyline}: {key} is not used")
            exit_code = 1
    return exit_code


def compare_action(args):
    exit_code = 0
    reference_st = StringTable.from_file(args.reference)
    source_st = StringTable.from_file(args.source)

    comp = StringTable.statistic(reference_st, source_st)

    if not args.summary_only:
        for key in source_st:
            if key in comp.right_only:
                print(
                    f"{comp.right.fpath}:{comp.right[key].keyline}: Key '{key}' is not in reference file {comp.left.fpath}"
                )
                exit_code = 1

            if key in comp.left_pure_reference:
                print(
                    f"{comp.right.fpath}:{comp.right[key].keyline}:"
                    f" Key '{key}' contains translation for pure reference in reference file"
                    f" value {comp.left.fpath}:{comp.left[key].keyline}"
                )
                exit_code = 1

            if key in comp.layout_mismatch:
                print(
                    f"{comp.right.fpath}:{comp.right[key].keyline}:"
                    f" Value of key '{key}' layout does not match that of reference file"
                    f" value {comp.left.fpath}:{comp.left[key].keyline}"
                )
                exit_code = 1

            if key in comp.right_older:
                print(
                    f"{comp.right.fpath}:{comp.right[key].keyline}:"
                    f" Value of key '{key}' is older than reference file"
                    f" value {comp.left.fpath}:{comp.left[key].keyline}"
                )
                exit_code = 1

            if key in comp.untranslated:
                print(
                    f"{comp.right.fpath}:{comp.right[key].keyline}:"
                    f" Value of key '{key}' has no translation compared to reference file"
                    f" {comp.left.fpath}:{comp.left[key].keyline}"
                )
                exit_code = 1

            if key in comp.identical:
                print(
                    f"{comp.right.fpath}:{comp.right[key].keyline}:"
                    f" Value of key '{key}' is identical to reference file"
                    f" {comp.left.fpath}:{comp.left[key].keyline}"
                )
            exit_code = 1

    print(
        f"""
Summary comparing '{comp.right.fpath}' against '{comp.left.fpath}':
    Keys matching - {len(comp.right) - len(comp.right_only)}/{len(comp.left)} ({100.0 * (len(comp.right) - len(comp.right_only)) / len(comp.left):3.1f}%)
    Keys not in reference - {len(comp.right_only)}
    Value is reference - {len(comp.left_pure_reference)}
    Values layout mismatch - {len(comp.layout_mismatch)}
    Values older than reference - {len(comp.right_older)}
    Values untranslated - {len(comp.untranslated)}
    Values same as reference - {len(comp.identical)}
""".strip()
    )
    return exit_code


if __name__ == "__main__":
    root_parser = argparse.ArgumentParser(description="Verify and modify string tables")
    verb_parsers = root_parser.add_subparsers(
        title="verbs",
        description=f"For more details run `{root_parser.prog} <verb> --help`.",
    )

    format_parser = verb_parsers.add_parser(
        "format",
        help="format a string table and exit",
        description=textwrap.dedent(
            """\
        Pretty prints a given string table onto the standard output.

        Formatting a string tables makes sure that:
        * Translation entries (translated or not) are separated by one empty
          line.
        * Translation notes (prefix: '# ') and untranslated entries
          (prefix: '#*') are properly prefixed and formatted
        * Translation notes, key and value don't have empty lines between
          each other.
        * String tables contain the language name as first line, followed by
          the file notes.
        * Section titles are prefixed with a block comment (prefix '##').
        * Section titles have two leading whitespaces and one trailing
          whitespace.
        """
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    format_parser.set_defaults(action=format_action)
    format_parser.add_argument(
        "source",
        metavar="SOURCE",
        help="string table to format",
        type=argparse.FileType(encoding="utf-8", errors="strict"),
    )

    sync_parser = verb_parsers.add_parser(
        "sync",
        help="synchronize two string tables and exit",
        description=textwrap.dedent(
            """\
        Synchronizes two string tables by copying over translation entry key,
        notes, section titles and includes from the reference into the source,
        while preserving existing translation values, language name and file notes
        from source.

        Source translation entries without matching reference counterpart are
        discarded.

        The resulting string table is printed out to standard output.
        """
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    sync_parser.set_defaults(action=sync_action)
    sync_parser.add_argument(
        "reference",
        metavar="REFERENCE",
        help="reference string table",
        type=argparse.FileType(encoding="utf-8", errors="strict"),
    )
    sync_parser.add_argument(
        "source",
        metavar="SOURCE",
        help="string table to sync",
        type=argparse.FileType(encoding="utf-8", errors="strict"),
    )

    rename_key_parser = verb_parsers.add_parser(
        "rename-key",
        help="rename all occurences of a key within a stringtable and exit",
        description=textwrap.dedent(
            """\
        Replace all occurances of a translation entry key within the given key,
        including internal references.

        The resulting string table is printed out to standard output.
        """
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    rename_key_parser.set_defaults(action=rename_key_action)
    rename_key_parser.add_argument(
        "source",
        metavar="SOURCE",
        help="string table to rename old key within",
        type=argparse.FileType(encoding="utf-8", errors="strict"),
    )
    rename_key_parser.add_argument("old_key", metavar="OLD_KEY", help="key to rename")
    rename_key_parser.add_argument("new_key", metavar="NEW_KEY", help="new key name")

    edit_parser = verb_parsers.add_parser(
        "edit",
        help="edit a stringtable and exit",
        description=textwrap.dedent(
            """\
        Starts a web server and a web browser to provide an interactive editor.

        Use ^C to exit the web server.  The string table will not be saved, so
        download it before exiting the web server or closing the browser.
        """
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    edit_parser.set_defaults(action=edit_action)
    edit_parser.add_argument(
        "reference",
        metavar="REFERENCE",
        help="reference string table",
        type=argparse.FileType(encoding="utf-8", errors="strict"),
    )
    edit_parser.add_argument(
        "source",
        metavar="SOURCE",
        help="string table to edit",
        type=argparse.FileType(encoding="utf-8", errors="strict"),
    )

    check_parser = verb_parsers.add_parser(
        "check",
        help="check a stringtable for consistency and exit",
        description=textwrap.dedent(
            """\
        Check if all references within translation entries are provided by the source or
        or the reference string table.
        """
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    check_parser.set_defaults(action=check_action)
    check_parser.add_argument(
        "-r",
        "--reference",
        metavar="REFERENCE",
        help="reference string table",
        type=argparse.FileType(encoding="utf-8", errors="strict"),
    )
    check_parser.add_argument(
        "sources",
        metavar="SOURCES",
        help="string tables to check",
        nargs="+",
        type=argparse.FileType(encoding="utf-8", errors="strict"),
    )

    compare_parser = verb_parsers.add_parser(
        "compare",
        help="compare two string tables and exit",
        description=textwrap.dedent(
            """\
        Compare two string tables and point out differences between them.
        """
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    compare_parser.set_defaults(action=compare_action)
    compare_parser.add_argument(
        "-s", "--summary-only", help="print only a summary of differences", action="store_true", dest="summary_only"
    )
    compare_parser.add_argument(
        "reference",
        metavar="REFERENCE",
        help="reference string table",
        type=argparse.FileType(encoding="utf-8", errors="strict"),
    )
    compare_parser.add_argument(
        "source",
        metavar="SOURCE",
        help="string table to compare",
        type=argparse.FileType(encoding="utf-8", errors="strict"),
    )

    args = root_parser.parse_args()
    if "action" not in args:
        root_parser.print_usage()
        root_parser.exit()
    sys.exit(args.action(args))



import os
import sys
from pathlib import Path
from shutil import rmtree

from check.focs_dump_parser._tools import get_base_dump_folder, get_branch
from check.focs_dump_parser.make_diff import report_diff
from check.focs_dump_parser.parser import parse_buildings


def is_windows():
    return sys.platform.startswith("win")


def get_log_file():
    if is_windows():
        app_data = Path(os.getenv("APPDATA"))
        return app_data / "FreeOrion/AI_1.log"
    else:
        raise NotImplementedError("Where logs are located?")


def main():
    branch = get_branch()

    dump_folder = get_base_dump_folder() / branch
    rmtree(dump_folder, ignore_errors=True)
    dump_folder.mkdir(parents=True, exist_ok=True)

    with get_log_file().open() as f:
        all_buildings = parse_buildings(f)

    for name, content in all_buildings:
        with open(dump_folder / f"{name}.txt", "w") as f:
            f.write(content)

    stats = f"Buildings parsed: {len(all_buildings)}"

    if branch != "master":
        diff_stats = report_diff()
    else:
        diff_stats = {}

    print(stats)
    for k, v in diff_stats.items():
        print(f"{k}({len(v)}): {','.join(v)}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import re
from copy import copy
from pathlib import Path
from sys import argv

re_tags = re.compile(r"^ *tags *= *\[")
re_start_effects = re.compile(r"^ *effectsgroups *= *\[")
re_start_foci = re.compile(r"^ *foci *= *\[")
re_end_list = re.compile(r"^ *\]")
re_macro = re.compile(r"^ *\[\[([A-Z_]+)\]\]")
issues_found = False
# Commented out are not used by the AI, most likely they never will.
# For tags that are implemented as species effects, only the AI
# is using the tags.
skills_to_check = [
    "INDUSTRY",
    "RESEARCH",
    "INFLUENCE",
    # "STOCKPILE",
    "POPULATION",
    "HAPPINESS",
    "SUPPLY",
    "OFFENSE_TROOPS",
    # "DEFENSE_TROOPS",
    "WEAPONS",
    "DETECTION",
    "STEALTH",
    "FUEL",
]
foci_to_check = [
    "INDUSTRY",
    "RESEARCH",
    "INFLUENCE",
]


def check_skill(name):
    if name.startswith(("AVERAGE", "PEDIA", "HAEMAESTHETIC", "INFINITE")):
        return ""
    for skill in skills_to_check:
        if name.endswith(skill):
            return skill
    return ""


def check_species(file):  # noqa: C901
    def issue(text):
        global issues_found
        issues_found = True
        print(f"{file.name}: {text}")

    effects = False
    foci = False
    tags = set()
    foci_list = set(foci_to_check)
    with open(file, encoding="utf-8") as f:
        for line in f.readlines():
            if re_tags.match(line):
                strings = line.split('"')
                for i in range(1, len(strings), 2):
                    if check_skill(strings[i]):
                        tags.add(strings[i])
                alltags = copy(tags)
            if re_start_effects.match(line):
                effects = True
            if re_start_foci.match(line):
                foci = True
            if effects or foci:
                r = re_macro.match(line)
                if r:
                    value = r.groups()[0]
                    if effects:
                        skill = check_skill(value)
                        if skill:
                            if value in tags:
                                tags.remove(value)
                            else:
                                issue("%s in effectgroups, missing in tags" % value)
                    else:
                        focus = value.split("_")[1]
                        if focus in foci_list:
                            foci_list.remove(focus)
                            if ("NO_" + focus) in alltags:
                                issue(f"{value}, but not tag NO_{focus}")
                elif re_end_list.match(line):
                    effects = False
                    foci = False
    for focus in foci_list:
        if ("NO_" + focus) not in alltags:
            issue(f"Not HAS_{focus}_FOCUS, NO_{focus} missing in tags")
    if tags:
        for tag in tags:
            issue("%s in tags, missing in effectgroups" % tag)


if __name__ == "__main__":
    if len(argv) != 2:
        print("Usage: %s <species file|directory contains species files>" % argv[0])
        exit(1)
    path = Path(argv[1])
    if path.is_dir():
        for f in path.iterdir():
            if f.name.startswith("SP_") and f.name.endswith(".txt"):
                check_species(f)
    else:
        check_species(path)
    exit(int(issues_found))



# // param1: value
# NATIVE_PLANETARY_DEFENSE
# '''EffectsGroup
#             scope = Source
#             activation = And [
#                 Planet
#                 Unowned
#             ]
#             accountinglabel = "NATIVE_PLANETARY_DEFENSE_LABEL"
#             effects = SetMaxDefense value = Value + ( @1@ )
# '''

from macros.misc import PLANET_DEFENSE_FACTOR
from macros.priorities import (
    AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
    TARGET_AFTER_2ND_SCALING_PRIORITY,
)

try:
    from focs._effects import (
        CurrentTurn,
        EffectsGroup,
        Focus,
        GameRule,
        IsSource,
        LocalCandidate,
        MinOf,
        NamedReal,
        Planet,
        SetDefense,
        SetMaxDefense,
        SetTargetHappiness,
        Target,
        Unowned,
        Value,
    )
except ModuleNotFoundError:
    pass


STANDARD_DEFENSE_GROWTH = EffectsGroup(  # increase 1 per turn, up to max
    scope=IsSource,
    activation=Planet() & Unowned & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn),
    accountinglabel="DEF_ROOT_DEFENSE",
    priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
    effects=SetDefense(value=MinOf(float, Value(Target.MaxDefense), Value + PLANET_DEFENSE_FACTOR)),
)


PROTECTION_FOCUS_DEFENSE = EffectsGroup(
    scope=IsSource,
    activation=Planet() & Focus(type=["FOCUS_PROTECTION"]),
    stackinggroup="FOCUS_PROTECTION_DEFENSE_STACK",
    accountinglabel="FOCUS_PROTECTION_LABEL",
    priority=TARGET_AFTER_2ND_SCALING_PRIORITY,
    effects=[
        SetMaxDefense(value=Value * 2),
        SetTargetHappiness(
            value=Value
            + NamedReal(
                name="PROTECION_FOCUS_STABILITY_BONUS",
                value=GameRule(type=float, name="RULE_PROTECTION_FOCUS_STABILITY"),
            )
        ),
    ],
)


AVERAGE_PLANETARY_DEFENSE = [
    PROTECTION_FOCUS_DEFENSE,
    STANDARD_DEFENSE_GROWTH,
]


# #include "../macros/priorities.macros"



from macros.priorities import (
    POPULATION_FIRST_PRIORITY,
    TARGET_POPULATION_AFTER_SCALING_PRIORITY,
    TARGET_POPULATION_BEFORE_SCALING_PRIORITY,
    TARGET_POPULATION_LAST_BEFORE_OVERRIDE_PRIORITY,
    TARGET_POPULATION_OVERRIDE_PRIORITY,
    TARGET_POPULATION_SCALING_PRIORITY,
)

try:
    from focs._effects import (
        Abs,
        AnyEmpire,
        BlackHole,
        Blue,
        Contains,
        CurrentTurn,
        EffectsGroup,
        EmpireHasAdoptedPolicy,
        Focus,
        GasGiantType,
        Good,
        HasSpecies,
        HasTag,
        Homeworld,
        Hostile,
        IsSource,
        IsTarget,
        LocalCandidate,
        MaxOf,
        MinOf,
        Neutron,
        NoStar,
        Orange,
        OwnedBy,
        Planet,
        Poor,
        Red,
        ResourceSupplyConnected,
        SetPopulation,
        SetStarType,
        SetTargetPopulation,
        Source,
        Star,
        StatisticIf,
        Target,
        TargetPopulation,
        Turn,
        Uninhabitable,
        Value,
        White,
        Yellow,
    )
except ModuleNotFoundError:
    pass

from species.species_macros.advanced_focus import ADVANCED_FOCUS_EFFECTS
from species.species_macros.general import (
    FOCUS_CHANGE_PENALTY,
    STANDARD_CONSTRUCTION,
    STANDARD_METER_GROWTH,
)

GROWTH_RATE_FACTOR = (
    0.1
    * (
        1 - StatisticIf(float, condition=IsTarget & EmpireHasAdoptedPolicy(name="PLC_NO_GROWTH"))
    )  # no growth with no-growth policy
    * (
        1 + 0.5 * StatisticIf(float, condition=IsTarget & EmpireHasAdoptedPolicy(name="PLC_POPULATION"))
    )  # +50% growth with population policy
    * (
        1
        + StatisticIf(float, condition=IsTarget & EmpireHasAdoptedPolicy(name="PLC_AUGMENTATION"))
        # slower growth with augmentation on low-infrastructure planets
        * MinOf(
            float,
            1.2,
            MaxOf(
                float,
                0.5,
                Target.Construction / 40.0,
            ),
        )
    )
)

HOMEWORLD_BONUS_POPULATION = EffectsGroup(
    scope=IsSource & Homeworld(name=[Source.Species]),
    activation=Planet(),
    stackinggroup="HOMEWORLD_STACK",
    accountinglabel="HOMEWORLD_BONUS",
    priority=TARGET_POPULATION_AFTER_SCALING_PRIORITY,
    effects=SetTargetPopulation(value=Value + 2 * Target.HabitableSize),
)

# Implements environmental modifiers from Growth techs.
# Changes to the growth tree should take this into account.

ENVIRONMENT_MODIFIER = [
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Planet(environment=[Uninhabitable]),
        accountinglabel="UNINHABTIABLE_ENVIRONMENT_LABEL",
        priority=TARGET_POPULATION_OVERRIDE_PRIORITY,
        effects=SetTargetPopulation(value=-999),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Planet(environment=[Hostile]),
        accountinglabel="HOSTILE_ENVIRONMENT_LABEL",
        priority=TARGET_POPULATION_BEFORE_SCALING_PRIORITY,
        effects=SetTargetPopulation(value=Value - 4 * Source.HabitableSize),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Planet(environment=[Poor]),
        accountinglabel="POOR_ENVIRONMENT_LABEL",
        priority=TARGET_POPULATION_BEFORE_SCALING_PRIORITY,
        effects=SetTargetPopulation(value=Value - 2 * Source.HabitableSize),
    ),
    #           EffectsGroup
    #             scope = Source
    #             activation = And [
    #                 Planet
    #                 Planet environment = Adequate
    #             ]
    #             accountinglabel = "ADEQUATE_ENVIRONMENT_LABEL"
    #             priority = [[TARGET_POPULATION_BEFORE_SCALING_PRIORITY]]
    #             effects = SetTargetPopulation value = Value + 0 * Source.HabitableSize
    EffectsGroup(
        scope=IsSource,
        activation=Planet(environment=[Good]),
        accountinglabel="GOOD_ENVIRONMENT_LABEL",
        priority=TARGET_POPULATION_BEFORE_SCALING_PRIORITY,
        effects=SetTargetPopulation(value=Value + 3 * Target.HabitableSize),
    ),
]

# This is dependent on current placement in population effects calc,
# just after Homeworld and Environment
SELF_SUSTAINING_BONUS = EffectsGroup(
    scope=IsSource,
    activation=Planet(environment=[Good]) & HasTag(name="SELF_SUSTAINING"),
    accountinglabel="SELF_SUSTAINING_LABEL",
    priority=TARGET_POPULATION_AFTER_SCALING_PRIORITY,
    effects=SetTargetPopulation(value=Value + 3 * Target.HabitableSize),  # Gets the same bonus as three growth specials
)

PHOTOTROPHIC_BONUS = [
    EffectsGroup(
        scope=Contains(IsSource),
        activation=Planet()
        & (
            OwnedBy(affiliation=AnyEmpire)
            | Star(type=[BlackHole, NoStar])  # Natives are allowed to have bright or dim stars
        )
        & HasTag(name="PHOTOTROPHIC")
        & Turn(high=0)
        & ~Star(type=[Orange]),
        effects=SetStarType(type=Yellow),  # start with a normal star to be balanced
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & HasTag(name="PHOTOTROPHIC") & Star(type=[Blue]) & TargetPopulation(low=0),
        accountinglabel="VERY_BRIGHT_STAR",
        priority=TARGET_POPULATION_LAST_BEFORE_OVERRIDE_PRIORITY,
        effects=SetTargetPopulation(value=Value + 3 * Source.HabitableSize),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & HasTag(name="PHOTOTROPHIC") & Star(type=[White]) & TargetPopulation(low=0),
        accountinglabel="BRIGHT_STAR",
        priority=TARGET_POPULATION_LAST_BEFORE_OVERRIDE_PRIORITY,
        effects=SetTargetPopulation(value=Value + 1.5 * Source.HabitableSize),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & HasTag(name="PHOTOTROPHIC") & Star(type=[Red, Neutron]),
        accountinglabel="DIM_STAR",
        priority=TARGET_POPULATION_LAST_BEFORE_OVERRIDE_PRIORITY,
        effects=SetTargetPopulation(value=Value - 1 * Source.HabitableSize),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & HasTag(name="PHOTOTROPHIC") & Star(type=[BlackHole, NoStar]),
        accountinglabel="NO_STAR",
        priority=TARGET_POPULATION_LAST_BEFORE_OVERRIDE_PRIORITY,
        effects=SetTargetPopulation(value=Value - 10 * Source.HabitableSize),
    ),
]

GASEOUS_BONUS = EffectsGroup(
    scope=IsSource,
    activation=Planet(type=[GasGiantType]) & HasTag(name="GASEOUS"),
    accountinglabel="GASEOUS_LABEL",
    priority=TARGET_POPULATION_SCALING_PRIORITY,
    effects=SetTargetPopulation(value=Value - 0.5 * Abs(float, Value)),
)

HOMEWORLD_GROWTH_FOCUS_BOOST = EffectsGroup(
    scope=Planet()
    & OwnedBy(empire=Source.Owner)
    & HasSpecies(name=[Source.Species])
    & ~Homeworld(name=[Source.Species])
    & ResourceSupplyConnected(empire=Source.Owner, condition=IsSource),
    activation=Planet() & Focus(type=["FOCUS_GROWTH"]) & Homeworld(),
    stackinggroup="HOMEWORLD_STACK",
    accountinglabel="HOMEWORLD_SUPPLY",
    priority=TARGET_POPULATION_AFTER_SCALING_PRIORITY,
    effects=SetTargetPopulation(value=Value + 1 * Target.HabitableSize),
)


BASIC_POPULATION = [
    HOMEWORLD_BONUS_POPULATION,
    *ENVIRONMENT_MODIFIER,
    SELF_SUSTAINING_BONUS,
    *PHOTOTROPHIC_BONUS,
    GASEOUS_BONUS,
    HOMEWORLD_GROWTH_FOCUS_BOOST,
    # population growth or decay towards to target population
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & (LocalCandidate.LastTurnConquered < CurrentTurn),
        priority=POPULATION_FIRST_PRIORITY,
        effects=SetPopulation(
            # growth is GROWTH_RATE_FACTOR * population * (maximum + 1 - population) / maximum
            # which gives factor for population 1, best growth around half full and aproaching
            # factor again when almost full
            value=(
                (Value < Value(Target.TargetPopulation))
                * MinOf(
                    float,
                    Value(Target.TargetPopulation),
                    Value + Value * GROWTH_RATE_FACTOR * (1 + (1 - Value) / Value(Target.TargetPopulation)),
                )
            )
            # shrink if overpopulated
            + (
                (Value >= Value(Target.TargetPopulation))
                * MaxOf(float, Value(Target.TargetPopulation), Value - 0.1 * (Value - Value(Target.TargetPopulation)))
            )
        ),
    ),
    # Since all species have the same advanced focus effects and
    # infrastructure, the macros are stashed here, where they don't
    # need to be manually included in each species' macros.
    *FOCUS_CHANGE_PENALTY,
    *ADVANCED_FOCUS_EFFECTS,
    *STANDARD_CONSTRUCTION,
    *STANDARD_METER_GROWTH,
]

AVERAGE_POPULATION = BASIC_POPULATION

# EXTREMELY_BAD_POPULATION
# '''     [[BASIC_POPULATION]]
#
#         EffectsGroup
#             description = "EXTREMELY_BAD_POPULATION_DESC"
#             scope = Source
#             activation = Planet
#             accountinglabel = "EXTREMELY_BAD_POPULATION_LABEL"
#             priority = [[TARGET_POPULATION_SCALING_PRIORITY]]
#             effects = SetTargetPopulation value = Value -0.75*abs(Value)
# '''

# VERY_BAD_POPULATION
# '''     [[BASIC_POPULATION]]
#
#         EffectsGroup
#             description = "VERY_BAD_POPULATION_DESC"
#             scope = Source
#             activation = Planet
#             accountinglabel = "VERY_BAD_POPULATION_LABEL"
#             priority = [[TARGET_POPULATION_SCALING_PRIORITY]]
#             effects = SetTargetPopulation value = Value -0.5*abs(Value)
# '''

# BAD_POPULATION
# '''     [[BASIC_POPULATION]]
#
#         EffectsGroup
#             description = "BAD_POPULATION_DESC"
#             scope = Source
#             activation = Planet
#             accountinglabel = "BAD_POPULATION_LABEL"
#             priority = [[TARGET_POPULATION_SCALING_PRIORITY]]
#             effects = SetTargetPopulation value = Value -0.25*abs(Value)
# '''

# GOOD_POPULATION
# '''     [[BASIC_POPULATION]]

#         EffectsGroup
#             description = "GOOD_POPULATION_DESC"
#             scope = Source
#             activation = Planet
#             accountinglabel = "GOOD_POPULATION_LABEL"
#             priority = [[TARGET_POPULATION_SCALING_PRIORITY]]
#             effects = SetTargetPopulation value = Value +0.25*abs(Value)
# '''

# // reduces max population in systems with bright stars
# LIGHT_SENSITIVE
# '''
#         EffectsGroup
#             description = "LIGHT_SENSITIVE_DESC"
#             scope = Source
#             activation = And [
#                 Planet
#                 Star type = [Blue]
#             ]
#             accountinglabel = "VERY_BRIGHT_STAR"
#             priority = [[TARGET_POPULATION_AFTER_SCALING_PRIORITY]]
#             effects = SetTargetPopulation value = Value - 2 * Source.HabitableSize

#         EffectsGroup
#             scope = Source
#             activation = And [
#                 Planet
#                 Star type = [White]
#             ]
#             accountinglabel = "BRIGHT_STAR"
#             priority = [[TARGET_POPULATION_AFTER_SCALING_PRIORITY]]
#             effects = SetTargetPopulation value = Value - Source.HabitableSize
# '''


# // @1@ Species key
# LIFECYCLE_MANIP_POPULATION_EFFECTS
# '''
#         EffectsGroup
#             scope = And [
#                 Object id = Source.PlanetID
#                 Planet
#             ]
#             activation = And [
#                 Not OwnerHasTech name = "GRO_LIFECYCLE_MAN"
#                 Turn low = LocalCandidate.CreationTurn + 1 high = LocalCandidate.CreationTurn + 1
#             ]
#             priority = [[POPULATION_DEFAULT_PRIORITY]]
#             effects = [
#                 SetSpecies name = @1@
#                 SetPopulation value = 1
#             ]
#
#         EffectsGroup
#             scope = And [
#                 Object id = Source.PlanetID
#                 Planet
#             ]
#             activation = And [
#                 OwnerHasTech name = "GRO_LIFECYCLE_MAN"
#                 Turn low = LocalCandidate.CreationTurn + 1 high = LocalCandidate.CreationTurn + 1
#             ]
#             priority = [[POPULATION_DEFAULT_PRIORITY]]
#             effects = [
#                 SetSpecies name = @1@
#                 SetPopulation value = [[MIN_RECOLONIZING_SIZE]]
#             ]
# '''

# #include "macros/advanced_focus.macros"
# #include "macros/focus.macros"
# #include "macros/general.macros"



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_ENRG_FRIGATE",
    description="SHP_ENRG_FRIGATE_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=192 * TECH_COST_MULTIPLIER,
    researchturns=4,
    tags=["PEDIA_ENERGY_HULL_TECHS"],
    prerequisites=["SHP_FRC_ENRG_COMP"],
    unlock=[Item(type=UnlockShipHull, name="SH_ENERGY_FRIGATE")],
    graphic="icons/ship_hulls/energy_frigate_hull_small.png",
)
